import os
import torch
import numpy as np
import torch.nn as nn

from torch.utils.data import DataLoader
from typing import Dict, List, Tuple
from ..optimizer import OptunaOptimizer, TrialPruned
from .. import DEVICE, set_seed  # pyLOM/NN/__init__.py
from ... import pprint, cr  # pyLOM/__init__.py
from ...utils.errors import raiseWarning


class MLP(nn.Module):
    r"""
    Multi-layer perceptron model for regression tasks. The model is based on the PyTorch library `torch.nn` 
    (detailed documentation can be found at https://pytorch.org/docs/stable/nn.html).

    Args:
        input_size (int): Number of input features.
        output_size (int): Number of output features.
        n_layers (int): Number of hidden layers.
        hidden_size (int): Number of neurons in each hidden layer.
        p_dropouts (float, optional): Dropout probability for the hidden layers (default: ``0.0``).
        checkpoint_file (str, optional): Path to a checkpoint file to load the model from (default: ``None``).
        activation (torch.nn.Module, optional): Activation function to use (default: ``torch.nn.functional.relu``).
        device (torch.device, optional): Device to use (default: ``torch.device("cpu")``).
        seed (int, optional): Seed for reproducibility (default: ``None``).
        kwargs: Additional keyword arguments.
    """
    def __init__(
        self,
        input_size: int,
        output_size: int,
        n_layers: int,
        hidden_size: int,
        p_dropouts: float = 0.0,
        activation: torch.nn.Module = torch.nn.functional.relu,
        device: torch.device = DEVICE,
        seed: int = None,
        **kwargs: Dict,
    ):
        self.input_size = input_size
        self.output_size = output_size
        self.n_layers = n_layers
        self.hidden_size = hidden_size
        self.p_dropouts = p_dropouts
        self.activation = activation
        self.device = device

        super().__init__()
        if seed is not None:
            set_seed(seed)
        self.layers = nn.ModuleList()
        for i in range(n_layers):
            in_size = input_size if i == 0 else hidden_size
            out_size = hidden_size
            self.layers.append(nn.Linear(in_size, out_size))
            if p_dropouts > 0:
                self.layers.append(nn.Dropout(p_dropouts))
        self.oupt = nn.Linear(hidden_size, output_size)

        for layer in self.layers:
            if (isinstance(layer, nn.Linear) and i % 2 == 0): # Initialize only non-dropout linear layers
                nn.init.xavier_uniform_(layer.weight)
                nn.init.zeros_(layer.bias)
        nn.init.xavier_uniform_(self.oupt.weight)
        nn.init.zeros_(self.oupt.bias)

        self.to(self.device)
    
    def forward(self, x):
        for layer in self.layers:
            z = self.activation(layer(x))
            x = z
        z = self.oupt(x)
        return z
    
    @cr('MLP.fit')
    def fit(
        self,
        train_dataset: torch.utils.data.Dataset,
        eval_dataset: torch.utils.data.Dataset = None,
        epochs: int = 100,
        lr: float = 0.001,
        lr_gamma: float = 1,
        lr_scheduler_step: int = 1,
        loss_fn: torch.nn.Module = torch.nn.MSELoss(),
        optimizer_class: torch.optim.Optimizer = torch.optim.Adam,
        scheduler_class: torch.optim.lr_scheduler.LRScheduler = torch.optim.lr_scheduler.StepLR,
        print_rate_batch: int = 0,
        print_rate_epoch: int = 1,
        **kwargs,
    )-> Dict[str, List[float]]:
        r"""
        Fit the model to the training data. If eval_set is provided, the model will be evaluated on this set after each epoch. 
        
        Args:
            train_dataset (torch.utils.data.Dataset): Training dataset to fit the model.
            eval_dataset (torch.utils.data.Dataset): Evaluation dataset to evaluate the model after each epoch (default: ``None``).
            epochs (int, optional): Number of epochs to train the model (default: ``100``).
            lr (float, optional): Learning rate for the optimizer (default: ``0.001``).
            lr_gamma (float, optional): Multiplicative factor of learning rate decay (default: ``1``).
            lr_scheduler_step (int, optional): Number of epochs to decay the learning rate (default: ``1``).
            loss_fn (torch.nn.Module, optional): Loss function to optimize (default: ``torch.nn.MSELoss()``).
            optimizer_class (torch.optim.Optimizer, optional): Optimizer class to use (default: ``torch.optim.Adam``).
            scheduler_class (torch.optim.lr_scheduler._LRScheduler, optional): Learning rate scheduler class to use. If ``None``, no scheduler will be used (default: ``torch.optim.lr_scheduler.StepLR``).
            print_rate_batch (int, optional): Print loss every ``print_rate_batch`` batches (default: ``1``). If set to ``0``, no print will be done.
            print_rate_epoch (int, optional): Print loss every ``print_rate_epoch`` epochs (default: ``1``). If set to ``0``, no print will be done.
            kwargs (dict, optional): Additional keyword arguments to pass to the DataLoader. Can be used to set the parameters of the DataLoader (see PyTorch documentation at https://pytorch.org/docs/stable/data.html#torch.utils.data.DataLoader):
                
                - batch_size (int, optional): Batch size (default: ``32``).
                - shuffle (bool, optional): Shuffle the data (default: ``True``).
                - num_workers (int, optional): Number of workers to use (default: ``0``).
                - pin_memory (bool, optional): Pin memory (default: ``True``).

        Returns:
            Dict[str, List[float]]: Dictionary containing the training and evaluation losses.
        """
        dataloader_params = {
            "batch_size": 32,
            "shuffle": True,
            "num_workers": 0,
            "pin_memory": True,
        }

        if not hasattr(self, "train_dataloader"):
            for key in dataloader_params.keys():
                if key in kwargs:
                    dataloader_params[key] = kwargs[key]
            train_dataloader = DataLoader(train_dataset, **dataloader_params)
        
        eval_dataloader = DataLoader(eval_dataset, **dataloader_params) if eval_dataset is not None else None 

        if not hasattr(self, "optimizer"):
            self.optimizer = optimizer_class(self.parameters(), lr=lr)
        if not hasattr(self, "scheduler"):
            self.scheduler = scheduler_class(self.optimizer, step_size=lr_scheduler_step, gamma=lr_gamma) if scheduler_class is not None else None
        
        if hasattr(self, "checkpoint"):
            self.optimizer.load_state_dict(self.checkpoint["state"][0])
            if self.scheduler is not None and len(self.checkpoint["state"][1]) > 0:
                self.scheduler.load_state_dict(self.checkpoint["state"][1])
                self.scheduler.gamma = lr_gamma
                self.scheduler.step_size = lr_scheduler_step
            epoch_list = self.checkpoint["state"][2]
            train_loss_list = self.checkpoint["state"][3]
            test_loss_list = self.checkpoint["state"][4]
        else:
            epoch_list = []
            train_loss_list = []
            test_loss_list = []

        total_epochs = len(epoch_list)+epochs
        for epoch in range(1+len(epoch_list), 1+total_epochs):
            train_loss = 0.0
            self.train()
            for b_idx, batch in enumerate(train_dataloader):
                x_train, y_train = batch[0].to(self.device), batch[1].to(self.device)
                self.optimizer.zero_grad()
                oupt = self(x_train)
                loss_val = loss_fn(oupt, y_train)
                loss_val.backward()
                self.optimizer.step()
                loss_val_item = loss_val.item()
                train_loss_list.append(loss_val_item)
                train_loss += loss_val_item
                if print_rate_batch != 0 and (b_idx % print_rate_batch) == 0:
                    pprint(0, "Batch %4d/%4d | Train loss (x1e5) %0.4f" % (b_idx, len(train_dataloader), loss_val_item * 1e5), flush=True)
                   
            train_loss = train_loss / (b_idx + 1)
            
            if self.scheduler is not None:
                self.scheduler.step()
            
            test_loss = 0.0
            if eval_dataloader is not None:
                self.eval()
                with torch.no_grad():
                    for n_idx, sample in enumerate(eval_dataloader):
                        x_test, y_test = sample[0].to(self.device), sample[1].to(self.device)
                        test_output = self(x_test)
                        loss_val = loss_fn(test_output, y_test)
                        test_loss += loss_val.item()

                test_loss = test_loss / (n_idx + 1)
                test_loss_list.append(test_loss)
            
            if print_rate_epoch != 0 and (epoch % print_rate_epoch) == 0:
                test_log = f" | Test loss (x1e5) {test_loss * 1e5:.4f}" if eval_dataloader is not None else ""
                pprint(0, f"Epoch {epoch}/{total_epochs} | Train loss (x1e5) {train_loss * 1e5:.4f} {test_log}", flush=True)

            epoch_list.append(epoch)
            self.state = (
                self.optimizer.state_dict(),
                self.scheduler.state_dict() if self.scheduler is not None else {},
                epoch_list,
                train_loss_list,
                test_loss_list,
                )
            
        return {"train_loss": train_loss_list, "test_loss": test_loss_list}
    
    @cr('MLP.predict')
    def predict(
        self, 
        X: torch.utils.data.Dataset, 
        return_targets: bool = False,
        **kwargs,
    ):
        r"""
        Predict the target values for the input data. The dataset is loaded to a DataLoader with the provided keyword arguments. 
        The model is set to evaluation mode and the predictions are made using the input data. 
        To make a prediction from a torch tensor, use the `__call__` method directly.

        Args:
            X (torch.utils.data.Dataset): The dataset whose target values are to be predicted using the input data.
            rescale_output (bool): Whether to rescale the output with the scaler of the dataset (default: ``True``).
            kwargs (dict, optional): Additional keyword arguments to pass to the DataLoader. Can be used to set the parameters of the DataLoader (see PyTorch documentation at https://pytorch.org/docs/stable/data.html#torch.utils.data.DataLoader):
                
                - batch_size (int, optional): Batch size (default: ``256``).
                - shuffle (bool, optional): Shuffle the data (default: ``False``).
                - num_workers (int, optional): Number of workers to use (default: ``0``).
                - pin_memory (bool, optional): Pin memory (default: ``True``).
 
        Returns:
            Tuple [np.ndarray, np.ndarray]: The predictions and the true target values.
        """
        dataloader_params = {
            "batch_size": 256,
            "shuffle": False,
            "num_workers": 0,
            "pin_memory": True,
        }
        
        for key in dataloader_params.keys():
            if key in kwargs:
                dataloader_params[key] = kwargs[key]

        predict_dataloader = DataLoader(X, **dataloader_params)

        total_rows = len(predict_dataloader.dataset)
        num_columns = self.output_size
        all_predictions = np.empty((total_rows, num_columns))
        all_targets = np.empty((total_rows, num_columns))

        self.eval()
        start_idx = 0
        with torch.no_grad():
            for x, y in predict_dataloader:
                output = self(x.to(self.device))
                batch_size = x.size(0)
                end_idx = start_idx + batch_size
                all_predictions[start_idx:end_idx, :] = output.cpu().numpy()
                if return_targets:
                    all_targets[start_idx:end_idx, :] = y.cpu().numpy()
                start_idx = end_idx

        if return_targets:
            return all_predictions, all_targets
        else:
            return all_predictions

    def save(
        self, 
        path: str,
    ):
        r"""
        Save the model to a checkpoint file.

        Args:
            path (str): Path to save the model. It can be either a path to a directory or a file name. 
            If it is a directory, the model will be saved with a filename that includes the number of epochs trained.
        """
        checkpoint = {
            "input_size": self.input_size,
            "output_size": self.output_size,
            "n_layers": self.n_layers,
            "hidden_size": self.hidden_size,
            "p_dropouts": self.p_dropouts,
            "activation": self.activation,
            "device": self.device,
            "state_dict": self.state_dict(),
            "state": self.state,
        }
        
        if os.path.isdir(path):
            filename = "/trained_model_{:06d}".format(len(self.state[2])) + ".pth"
            path = path + filename
        torch.save(checkpoint, path)
    
    @classmethod
    def load(
        cls, 
        path: str,
        device: torch.device = DEVICE,
    ):
        r"""
        Load the model from a checkpoint file. Does not require the model to be instantiated.

        Args:
            path (str): Path to the file to load the model from.
            device (torch.device, optional): Device to use (default: ``torch.device("cpu")``).

        Returns:
            Model (MLP): The loaded model.
        """
        checkpoint = torch.load(path, map_location=device, weights_only=False)
        raiseWarning("The model has been loaded with weights_only set to False. According with torch documentation, this is not recommended if you do not trust the source of your saved model, as it could lead to arbitrary code execution.")
        checkpoint['device'] = device
        model = cls(
            checkpoint["input_size"],
            checkpoint["output_size"],
            checkpoint["n_layers"],
            checkpoint["hidden_size"],
            checkpoint["p_dropouts"],
            checkpoint["activation"],
            checkpoint["device"],
        )
        
        model.load_state_dict(checkpoint["state_dict"])
        model.checkpoint = checkpoint
        return model
    
    @classmethod
    @cr('MLP.create_optimized_model')
    def create_optimized_model(
        cls, 
        train_dataset: torch.utils.data.Dataset, 
        eval_dataset: torch.utils.data.Dataset, 
        optuna_optimizer: OptunaOptimizer,
        **kwargs,
    ) -> Tuple[nn.Module, Dict]:
        r"""
        Create an optimized model using Optuna. The model is trained on the training dataset and evaluated on the validation dataset.
        
        Args:
            train_dataset (torch.utils.data.Dataset): The training dataset.
            eval_dataset (torch.utils.data.Dataset): The evaluation dataset.
            optuna_optimizer (OptunaOptimizer): The optimizer to use for optimization.
            kwargs: Additional keyword arguments.

        Returns:
            Tuple [MLP, Dict]: The optimized model and the optimization parameters.

        Example:
            >>> from pyLOM.NN import MLP, OptunaOptimizer
            >>> # Split the dataset
            >>> train_dataset, eval_dataset = dataset.get_splits([0.8, 0.2])
            >>> 
            >>> # Define the optimization parameters
            >>> optimization_params = {
            >>>     "lr": (0.00001, 0.01), # optimizable parameter
            >>>     "epochs": 50, # fixed parameter
            >>>     "n_layers": (1, 4),
            >>>     "batch_size": (128, 512),
            >>>     "hidden_size": (200, 400),
            >>>     "p_dropouts": (0.1, 0.5),
            >>>     "num_workers": 0,
            >>>     'print_rate_epoch': 5
            >>> }
            >>>
            >>> # Define the optimizer
            >>> optimizer = OptunaOptimizer(
            >>>     optimization_params=optimization_params,
            >>>     n_trials=5,
            >>>     direction="minimize",
            >>>     pruner=optuna.pruners.MedianPruner(n_startup_trials=5, n_warmup_steps=5, interval_steps=1),
            >>>     save_dir=None,
            >>> )
            >>>
            >>> # Create the optimized model
            >>> model, optimization_params = MLP.create_optimized_model(train_dataset, eval_dataset, optimizer)
            >>> 
            >>> # Fit the model
            >>> model.fit(train_dataset, eval_dataset, **optimization_params)
        """
        optimization_params = optuna_optimizer.optimization_params
        input_dim, output_dim = train_dataset[0][0].shape[0], train_dataset[0][1].shape[0]
        def optimization_function(trial) -> float:
            training_params = {}       
            for key, params in optimization_params.items():
                training_params[key] = cls._get_optimizing_value(key, params, trial)
            model = cls(input_dim, output_dim, **training_params)
            if optuna_optimizer.pruner is not None:
                epochs = training_params["epochs"]
                training_params["epochs"] = 1
                for epoch in range(epochs):
                    model.fit(train_dataset, **training_params)
                    y_pred, y_true = model.predict(eval_dataset, return_targets=True)
                    loss_val = ((y_pred - y_true)**2).mean()
                    trial.report(loss_val, epoch)
                    if trial.should_prune(): 
                        raise TrialPruned()
            else:
                model.fit(train_dataset, **training_params)
                y_pred, y_true = model.predict(eval_dataset, return_targets=True)
                loss_val = ((y_pred - y_true)**2).mean()
            
            return loss_val
        
        best_params = optuna_optimizer.optimize(objective_function=optimization_function)

        # Update params with best ones
        for param in best_params.keys():
            if param in optimization_params:
                optimization_params[param] = best_params[param]
        
        return cls(input_dim, output_dim, **optimization_params), optimization_params

    def _get_optimizing_value(name, value, trial):
        if isinstance(value, tuple) or isinstance(value, list):
            use_log = value[1] / value[0] >= 1000
            if isinstance(value[0], int):
                return trial.suggest_int(name, value[0], value[1], log=use_log)
            elif isinstance(value[0], float):
                return trial.suggest_float(name, value[0], value[1], log=use_log)
        else:
            return value



#!/usr/bin/env python
#
# PYLOM Testsuite POD
#
# Last revision: 20/09/2024
from __future__ import print_function, division

import sys, os, json, numpy as np
import pyLOM


## Parameters
DATAFILE  = sys.argv[1]
VARIABLES = eval(sys.argv[2])
OUTDIR    = sys.argv[3]
PARAMS    = json.loads(str(sys.argv[4]).replace("'",'"').lower())

## Data loading
m = pyLOM.Mesh.load(DATAFILE)
d = pyLOM.Dataset.load(DATAFILE,ptable=m.partition_table)
X = d.X(*VARIABLES)
t = d.get_variable('time')


## Run POD
PSI,S,V = pyLOM.POD.run(X,**PARAMS['run']) # PSI are POD modes
if pyLOM.utils.is_rank_or_serial(root=0): 
	fig,_ = pyLOM.POD.plotResidual(S)
	os.makedirs(OUTDIR,exist_ok=True)
	fig.savefig(f'{OUTDIR}/residuals.png',dpi=300)
# Truncate according to a residual
PSI,S,V = pyLOM.POD.truncate(PSI,S,V,r=PARAMS['r_trunc'])
print("KKK",PSI.shape)
pyLOM.POD.save(f'{OUTDIR}/results.h5',PSI,S,V,d.partition_table,nvars=len(VARIABLES),pointData=d.point)
# Reconstruct the flow
X_POD = pyLOM.POD.reconstruct(PSI,S,V)
# Compute RMSE
rmse = pyLOM.math.RMSE(X_POD,X)


## Testsuite output
M = pyLOM.POD.extract_modes(PSI,1,len(d),modes=PARAMS['modes'])
pyLOM.pprint(0,'TSUITE RMSE  = %e'%rmse)
pyLOM.pprint(0,'TSUITE X     =',X.min(),X.max(),X.mean())
pyLOM.pprint(0,'TSUITE S     =',S.min(),S.max(),S.mean())
pyLOM.pprint(0,'TSUITE V     =',V.min(),V.max(),V.mean())
pyLOM.pprint(0,'TSUITE X_POD =',X_POD.min(),X_POD.max(),X_POD.mean())
pyLOM.pprint(0,'TSUITE modes =',M.min(),M.max(),M.mean())


## Dump to ParaView
# Spatial modes
d.add_field('spatial_modes',len(PARAMS['modes']),pyLOM.POD.extract_modes(PSI,1,len(d),modes=PARAMS['modes']))
pyLOM.io.pv_writer(m,d,'modes',basedir=f'{OUTDIR}/modes',instants=[0],times=[0.],vars=['spatial_modes'],fmt='vtkh5')

# Temporal evolution
d.add_field('RECON',len(VARIABLES),X_POD)
pyLOM.io.pv_writer(m,d,'flow',basedir=f'{OUTDIR}/flow',instants=np.arange(t.shape[0],dtype=np.int32),times=t,vars=VARIABLES+['RECON'],fmt='vtkh5')


## Plot POD mode
if pyLOM.utils.is_rank_or_serial(0):
	# 0 - module, 1,2 - components
	fig, _ = pyLOM.POD.plotMode(V[:,:-1],t[:-1],modes=PARAMS['modes'])
	for i,f in enumerate(fig): f.savefig(f'{OUTDIR}/modes_%d.png'%i,dpi=300)


## Show and print timings
pyLOM.cr_info()
pyLOM.pprint(0,'End of output')


#!/usr/bin/env python
#
# DMD analysis.
#
# Last revision: 29/10/2021
from __future__ import print_function, division

import mpi4py
mpi4py.rc.recv_mprobe = False

import numpy as np
import pyLOM


## Data loading
DATAFILE  = './channel.h5'
VARIABLES = ['VELOC']

m  = pyLOM.Mesh.load(DATAFILE)
d  = pyLOM.Dataset.load(DATAFILE,ptable=m.partition_table)
X  = d.X(*VARIABLES)[0:3*d.mesh.npoints:3,:].copy() # Select only X component
t  = d.get_variable('time')
dt = t[1] - t[0]


## Run DMD
muReal,muImag,Phi,bJov = pyLOM.DMD.run(X,1e-6,remove_mean=False)
# Compute frequency and damping ratio of the modes
delta, omega = pyLOM.DMD.frequency_damping(muReal,muImag,dt)
pyLOM.DMD.save('results.h5',muReal,muImag,Phi,bJov,d.partition_table,nvars=1,pointData=d.point)

# plots
if pyLOM.utils.is_rank_or_serial(0):
	fig,_ = pyLOM.DMD.ritzSpectrum(muReal,muImag)               # Ritz Spectrum
	fig.savefig('ritzSpectrum.png',dpi=300)
	fig,_ = pyLOM.DMD.amplitudeFrequency(omega,bJov,norm=False) # Amplitude vs frequency
	fig.savefig('amplitudeFrequency.png',dpi=300)
	fig,_ = pyLOM.DMD.dampingFrequency(omega,delta)             # Damping ratio vs frequency
	fig.savefig('dampingFrequency.png',dpi=300)

# Spatial modes
modes = np.arange(1,100+1,dtype=np.int32)
d.add_field('U_MODES_REAL',len(modes),pyLOM.DMD.extract_modes(Phi,1,len(d),modes=modes,real=True))
d.add_field('U_MODES_IMAG',len(modes),pyLOM.DMD.extract_modes(Phi,1,len(d),modes=modes,real=False))
pyLOM.io.pv_writer(m,d,'modes',basedir='modes',instants=[0],times=[0.],vars=['U_MODES_REAL','U_MODES_IMAG'],fmt='vtkh5')


## Prediction
t_new = time[-1] + dt*np.arange(0,100,1,np.double)
X_DMD = pyLOM.DMD.reconstruction_jovanovic(Phi,muReal,muImag,t_new,bJov)
d.add_field('VELXP',1,X_DMD)
pyLOM.io.pv_writer(m,d,'pred',basedir='flow',instants=np.arange(t_new.shape[0],dtype=np.int32),times=t_new,vars=['VELXP'],fmt='vtkh5')


## Show and print timings
pyLOM.cr_info()


#!/usr/bin/env cpython
#
# pyLOM - Python Low Order Modeling.
#
# Math operations module - statistics.
#
# Last rev: 27/10/2021
from __future__ import print_function, division

import numpy as np

from ..utils.gpu import cp
from ..utils     import cr_nvtx as cr, mpi_reduce


@cr('math.RMSE')
def RMSE(A,B,relative=True):
	'''
	Compute RMSE between A and B
	'''
	p = cp if type(A) is cp.ndarray else np
	diff  = (A-B)
	sum1g = mpi_reduce(p.sum(diff*diff),op='sum',all=True)
	sum2g = mpi_reduce(p.sum(A*A),op='sum',all=True) if relative else p.prod(mpi_reduce(p.array(A.shape),op='sum',all=True))
	rmse  = np.sqrt(sum1g/sum2g)
	return rmse

@cr('math.MAE')
def MAE(A,B):
	'''
	Compute MAE between A and B
	'''
	p = cp if type(A) is cp.ndarray else np
	diff  = p.abs(A-B)
	sum1g = mpi_reduce(p.sum(diff),op='sum',all=True)
	sum2g = p.prod(mpi_reduce(p.array(A.shape),op='sum',all=True))
	mae   = sum1g/sum2g
	return mae

@cr('math.r2')
def r2(A,B):
	'''
	Compute r2 score between A and B
	'''
	p = cp if type(A) is cp.ndarray else np
	num  = (A-B)
	numg = mpi_reduce(p.sum(num*num),op='sum',all=True)
	sumg = mpi_reduce(p.sum(A),op='sum',all=True)
	sum2g = p.prod(mpi_reduce(p.array(A.shape),op='sum',all=True))
	den  = A - sumg/sum2g
	deng = mpi_reduce(p.sum(den*den),op='sum',all=True)
	r2   = 1. - numg/deng
	return r2


#!/usr/bin/env python
#
# pyLOM, utils.
#
# GPU setup routines
#
# Last rev: 14/02/2025
from __future__ import print_function, division

import numpy as np

from .mpi    import MPI_RANK
from .errors import raiseWarning

try:
	import cupy as cp

	def gpu_device(id=MPI_RANK,gpu_per_node=4):
		'''
		Setup the GPU to be used
		'''
		local_id = int(id%gpu_per_node)
		cp.cuda.Device(local_id).use()

	def gpu_to_cpu(X):
		'''
		Move an array from GPU to CPU
		'''
		return cp.asnumpy(X) if type(X) is cp.ndarray else X

	def cpu_to_gpu(X):
		'''
		Move an array from GPU to CPU
		'''
		return cp.asarray(X) if type(X) is not cp.ndarray else X
	
	def ascontiguousarray(X):
		p = cp if type(X) is cp.ndarray else np
		return p.ascontiguousarray(X)

except:
	import numpy as cp

	def gpu_device(id=MPI_RANK,gpu_per_node=4):
		'''
		Setup the GPU to be used
		'''
		raiseWarning('cupy not available! GPU version deactivated!')

	def gpu_to_cpu(X):
		'''
		Move an array from GPU to CPU
		'''
		return X

	def cpu_to_gpu(X):
		'''
		Move an array from GPU to CPU
		'''
		return X

	def ascontiguousarray(X):
		return np.ascontiguousarray(X)


#!/usr/bin/env cpython
#
# pyLOM - Python Low Order Modeling.
#
# Math operations module.
#
# Last rev: 27/10/2021

cimport cython
cimport numpy as np

import numpy as np

from libc.time     cimport time
from libc.stdlib   cimport malloc, free
from libc.string   cimport memcpy, memset
from .cfuncs       cimport real, real_complex, real_full
from .cfuncs       cimport c_sqr, c_ssvd, c_stsqr, c_stsqr_svd, c_srandomized_qr, c_sinit_randomized_qr, c_supdate_randomized_qr, c_srandomized_svd
from .cfuncs       cimport c_dqr, c_dsvd, c_dtsqr, c_dtsqr_svd, c_drandomized_qr, c_dinit_randomized_qr, c_dupdate_randomized_qr, c_drandomized_svd
from .cfuncs       cimport c_cqr, c_csvd, c_ctsqr, c_ctsqr_svd
from .cfuncs       cimport c_zqr, c_zsvd, c_ztsqr, c_ztsqr_svd
from ..utils.cr     import cr
from ..utils.errors import raiseError


@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def _sqr(float[:,:] A):
	'''
	QR factorization using Lapack
		Q(m,n) is the Q matrix
		R(n,n) is the R matrix
	'''
	cdef int retval, m = A.shape[0], n = A.shape[1]
	cdef np.ndarray[np.float32_t,ndim=2] Q = np.zeros((m,n),dtype=np.float32)
	cdef np.ndarray[np.float32_t,ndim=2] R = np.zeros((n,n),dtype=np.float32)
	retval = c_sqr(&Q[0,0],&R[0,0],&A[0,0],m,n)
	if not retval == 0: raiseError('Problems computing QR factorization!')
	return Q,R

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def _dqr(double[:,:] A):
	'''
	QR factorization using Lapack
		Q(m,n) is the Q matrix
		R(n,n) is the R matrix
	'''
	cdef int retval, m = A.shape[0], n = A.shape[1]
	cdef np.ndarray[np.double_t,ndim=2] Q = np.zeros((m,n),dtype=np.double)
	cdef np.ndarray[np.double_t,ndim=2] R = np.zeros((n,n),dtype=np.double)
	retval = c_dqr(&Q[0,0],&R[0,0],&A[0,0],m,n)
	if not retval == 0: raiseError('Problems computing QR factorization!')
	return Q,R

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def _cqr(np.complex64_t[:,:] A):
	'''
	QR factorization using Lapack
		Q(m,n) is the Q matrix
		R(n,n) is the R matrix
	'''
	cdef int retval, m = A.shape[0], n = A.shape[1]
	cdef np.ndarray[np.complex64_t,ndim=2] Q = np.zeros((m,n),dtype=np.complex64)
	cdef np.ndarray[np.complex64_t,ndim=2] R = np.zeros((n,n),dtype=np.complex64)
	retval = c_cqr(&Q[0,0],&R[0,0],&A[0,0],m,n)
	if not retval == 0: raiseError('Problems computing QR factorization!')
	return Q,R

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def _zqr(np.complex128_t[:,:] A):
	'''
	QR factorization using Lapack
		Q(m,n) is the Q matrix
		R(n,n) is the R matrix
	'''
	cdef int retval, m = A.shape[0], n = A.shape[1]
	cdef np.ndarray[np.complex128_t,ndim=2] Q = np.zeros((m,n),dtype=np.complex128)
	cdef np.ndarray[np.complex128_t,ndim=2] R = np.zeros((n,n),dtype=np.complex128)
	retval = c_zqr(&Q[0,0],&R[0,0],&A[0,0],m,n)
	if not retval == 0: raiseError('Problems computing QR factorization!')
	return Q,R

@cr('math.qr')
@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def qr(real_full[:,:] A):
	'''
	QR factorization using Lapack
		Q(m,n) is the Q matrix
		R(n,n) is the R matrix
	'''
	if real_full is np.complex128_t:
		return _zqr(A)
	elif real_full is np.complex64_t:
		return _cqr(A)
	elif real_full is double:
		return _dqr(A)
	else:
		return _sqr(A)

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def _ssvd(float[:,:] A, int do_copy):
	'''
	Single value decomposition (SVD) using Lapack.
		U(m,n)   are the POD modes.
		S(n)     are the singular values.
		V(n,n)   are the right singular vectors.
	'''
	cdef int retval
	cdef int m = A.shape[0], n = A.shape[1], mn = min(m,n)
	cdef float *Y_copy
	cdef np.ndarray[np.float32_t,ndim=2] U = np.zeros((m,mn),dtype=np.float32)
	cdef np.ndarray[np.float32_t,ndim=1] S = np.zeros((mn,) ,dtype=np.float32)
	cdef np.ndarray[np.float32_t,ndim=2] V = np.zeros((n,mn),dtype=np.float32)
	# Compute SVD
	if do_copy:
		Y_copy = <float*>malloc(m*n*sizeof(float))
		memcpy(Y_copy,&A[0,0],m*n*sizeof(float))
		retval = c_ssvd(&U[0,0],&S[0],&V[0,0],Y_copy,m,n)
		free(Y_copy)
	else:
		retval = c_ssvd(&U[0,0],&S[0],&V[0,0],&A[0,0],m,n)
	if not retval == 0: raiseError('Problems computing SVD!')
	return U,S,V

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def _dsvd(double[:,:] A, int do_copy):
	'''
	Single value decomposition (SVD) using Lapack.
		U(m,n)   are the POD modes.
		S(n)     are the singular values.
		V(n,n)   are the right singular vectors.
	'''
	cdef int retval
	cdef int m = A.shape[0], n = A.shape[1], mn = min(m,n)
	cdef double *Y_copy
	cdef np.ndarray[np.double_t,ndim=2] U = np.zeros((m,mn),dtype=np.double)
	cdef np.ndarray[np.double_t,ndim=1] S = np.zeros((mn,) ,dtype=np.double)
	cdef np.ndarray[np.double_t,ndim=2] V = np.zeros((n,mn),dtype=np.double)
	# Compute SVD
	if do_copy:
		Y_copy = <double*>malloc(m*n*sizeof(double))
		memcpy(Y_copy,&A[0,0],m*n*sizeof(double))
		retval = c_dsvd(&U[0,0],&S[0],&V[0,0],Y_copy,m,n)
		free(Y_copy)
	else:
		retval = c_dsvd(&U[0,0],&S[0],&V[0,0],&A[0,0],m,n)
	if not retval == 0: raiseError('Problems computing SVD!')
	return U,S,V

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def _csvd(np.complex64_t[:,:] A, int do_copy):
	'''
	Single value decomposition (SVD) using Lapack.
		U(m,n)   are the POD modes.
		S(n)     are the singular values.
		V(n,n)   are the right singular vectors.
	'''
	cdef int retval
	cdef int m = A.shape[0], n = A.shape[1], mn = min(m,n)
	cdef np.complex64_t *Y_copy
	cdef np.ndarray[np.complex64_t,ndim=2] U = np.zeros((m,mn),dtype=np.complex64)
	cdef np.ndarray[np.float32_t,ndim=1]     S = np.zeros((mn,) ,dtype=np.float32)
	cdef np.ndarray[np.complex64_t,ndim=2] V = np.zeros((n,mn),dtype=np.complex64)
	# Compute SVD
	if do_copy:
		Y_copy = <np.complex64_t*>malloc(m*n*sizeof(np.complex64_t))
		memcpy(Y_copy,&A[0,0],m*n*sizeof(np.complex64_t))
		retval = c_csvd(&U[0,0],&S[0],&V[0,0],Y_copy,m,n)
		free(Y_copy)
	else:
		retval = c_csvd(&U[0,0],&S[0],&V[0,0],&A[0,0],m,n)
	if not retval == 0: raiseError('Problems computing SVD!')
	return U,S,V

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def _zsvd(np.complex128_t[:,:] A, int do_copy):
	'''
	Single value decomposition (SVD) using Lapack.
		U(m,n)   are the POD modes.
		S(n)     are the singular values.
		V(n,n)   are the right singular vectors.
	'''
	cdef int retval
	cdef int m = A.shape[0], n = A.shape[1], mn = min(m,n)
	cdef np.complex128_t *Y_copy
	cdef np.ndarray[np.complex128_t,ndim=2] U = np.zeros((m,mn),dtype=np.complex128)
	cdef np.ndarray[np.double_t,ndim=1]     S = np.zeros((mn,) ,dtype=np.double)
	cdef np.ndarray[np.complex128_t,ndim=2] V = np.zeros((n,mn),dtype=np.complex128)
	# Compute SVD
	if do_copy:
		Y_copy = <np.complex128_t*>malloc(m*n*sizeof(np.complex128_t))
		memcpy(Y_copy,&A[0,0],m*n*sizeof(np.complex128_t))
		retval = c_zsvd(&U[0,0],&S[0],&V[0,0],Y_copy,m,n)
		free(Y_copy)
	else:
		retval = c_zsvd(&U[0,0],&S[0],&V[0,0],&A[0,0],m,n)
	if not retval == 0: raiseError('Problems computing SVD!')
	return U,S,V

@cr('math.svd')
@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def svd(real_full[:,:] A, int do_copy=True):
	'''
	Single value decomposition (SVD) using Lapack.
		U(m,n)   are the POD modes.
		S(n)     are the singular values.
		V(n,n)   are the right singular vectors.
	'''
	if real_full is np.complex128_t:
		return _zsvd(A,do_copy)
	elif real_full is np.complex64_t:
		return _csvd(A,do_copy)
	elif real_full is double:
		return _dsvd(A,do_copy)	
	else:
		return _ssvd(A,do_copy)

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def _stsqr(float[:,:] A):
	'''
	Parallel QR factorization using Lapack
		Q(m,n) is the Q matrix
		R(n,n) is the R matrix
	'''
	cdef int retval
	cdef int m = A.shape[0], n = A.shape[1]
	cdef np.ndarray[np.float32_t,ndim=2] Qi = np.zeros((m,n),dtype=np.float32)
	cdef np.ndarray[np.float32_t,ndim=2] R  = np.zeros((n,n),dtype=np.float32)
	# Compute SVD using TSQR algorithm
	retval = c_stsqr(&Qi[0,0],&R[0,0],&A[0,0],m,n)
	if not retval == 0: raiseError('Problems computing TSQR!')
	return Qi,R

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def _dtsqr(double[:,:] A):
	'''
	Parallel QR factorization using Lapack
		Q(m,n) is the Q matrix
		R(n,n) is the R matrix
	'''
	cdef int retval
	cdef int m = A.shape[0], n = A.shape[1]
	cdef np.ndarray[np.double_t,ndim=2] Qi = np.zeros((m,n),dtype=np.double)
	cdef np.ndarray[np.double_t,ndim=2] R  = np.zeros((n,n),dtype=np.double)
	# Compute SVD using TSQR algorithm
	retval = c_dtsqr(&Qi[0,0],&R[0,0],&A[0,0],m,n)
	if not retval == 0: raiseError('Problems computing TSQR!')
	return Qi,R

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def _ctsqr(np.complex64_t[:,:] A):
	'''
	Parallel QR factorization using Lapack
		Q(m,n) is the Q matrix
		R(n,n) is the R matrix
	'''
	cdef int retval
	cdef int m = A.shape[0], n = A.shape[1]
	cdef np.ndarray[np.complex64_t,ndim=2] Qi = np.zeros((m,n),dtype=np.complex64)
	cdef np.ndarray[np.complex64_t,ndim=2] R  = np.zeros((n,n),dtype=np.complex64)
	# Compute SVD using TSQR algorithm
	retval = c_ctsqr(&Qi[0,0],&R[0,0],&A[0,0],m,n)
	if not retval == 0: raiseError('Problems computing TSQR!')
	return Qi,R

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def _ztsqr(np.complex128_t[:,:] A):
	'''
	Parallel QR factorization using Lapack
		Q(m,n) is the Q matrix
		R(n,n) is the R matrix
	'''
	cdef int retval
	cdef int m = A.shape[0], n = A.shape[1]
	cdef np.ndarray[np.complex128_t,ndim=2] Qi = np.zeros((m,n),dtype=np.complex128)
	cdef np.ndarray[np.complex128_t,ndim=2] R  = np.zeros((n,n),dtype=np.complex128)
	# Compute SVD using TSQR algorithm
	retval = c_ztsqr(&Qi[0,0],&R[0,0],&A[0,0],m,n)
	if not retval == 0: raiseError('Problems computing TSQR!')
	return Qi,R

@cr('math.tsqr')
@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def tsqr(real_full[:,:] A):
	'''
	Parallel QR factorization using Lapack
		Q(m,n) is the Q matrix
		R(n,n) is the R matrix
	'''
	if real_full is np.complex128_t:
		return _ztsqr(A)
	elif real_full is np.complex64_t:
		return _ctsqr(A)
	elif real_full is double:
		return _dtsqr(A)
	else:
		return _stsqr(A)

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def _stsqr_svd(float[:,:] A):
	'''
	Parallel Single value decomposition (SVD) using Lapack.
		U(m,n)   are the POD modes.
		S(n)     are the singular values.
		V(n,n)   are the right singular vectors.
	'''
	cdef int retval
	cdef int m = A.shape[0], n = A.shape[1], mn = min(m,n)
	cdef np.ndarray[np.float32_t,ndim=2] U = np.zeros((m,mn),dtype=np.float32)
	cdef np.ndarray[np.float32_t,ndim=1] S = np.zeros((mn,) ,dtype=np.float32)
	cdef np.ndarray[np.float32_t,ndim=2] V = np.zeros((n,mn),dtype=np.float32)
	# Compute SVD using TSQR algorithm
	retval = c_stsqr_svd(&U[0,0],&S[0],&V[0,0],&A[0,0],m,n)
	if not retval == 0: raiseError('Problems computing TSQR SVD!')
	return U,S,V

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def _dtsqr_svd(double[:,:] A):
	'''
	Parallel Single value decomposition (SVD) using Lapack.
		U(m,n)   are the POD modes.
		S(n)     are the singular values.
		V(n,n)   are the right singular vectors.
	'''
	cdef int retval
	cdef int m = A.shape[0], n = A.shape[1], mn = min(m,n)
	cdef np.ndarray[np.double_t,ndim=2] U = np.zeros((m,mn),dtype=np.double)
	cdef np.ndarray[np.double_t,ndim=1] S = np.zeros((mn,) ,dtype=np.double)
	cdef np.ndarray[np.double_t,ndim=2] V = np.zeros((n,mn),dtype=np.double)
	# Compute SVD using TSQR algorithm
	retval = c_dtsqr_svd(&U[0,0],&S[0],&V[0,0],&A[0,0],m,n)
	if not retval == 0: raiseError('Problems computing TSQR SVD!')
	return U,S,V

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def _ctsqr_svd(np.complex64_t[:,:] A):
	'''
	Parallel Single value decomposition (SVD) using Lapack.
		U(m,n)   are the POD modes.
		S(n)     are the singular values.
		V(n,n)   are the right singular vectors.
	'''
	cdef int retval
	cdef int m = A.shape[0], n = A.shape[1], mn = min(m,n)
	cdef np.ndarray[np.complex64_t,ndim=2] U = np.zeros((m,mn),dtype=np.complex64)
	cdef np.ndarray[np.float32_t,ndim=1]     S = np.zeros((mn,) ,dtype=np.float32)
	cdef np.ndarray[np.complex64_t,ndim=2] V = np.zeros((n,mn),dtype=np.complex64)
	# Compute SVD using TSQR algorithm
	retval = c_ctsqr_svd(&U[0,0],&S[0],&V[0,0],&A[0,0],m,n)
	if not retval == 0: raiseError('Problems computing TSQR SVD!')
	return U,S,V

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def _ztsqr_svd(np.complex128_t[:,:] A):
	'''
	Parallel Single value decomposition (SVD) using Lapack.
		U(m,n)   are the POD modes.
		S(n)     are the singular values.
		V(n,n)   are the right singular vectors.
	'''
	cdef int retval
	cdef int m = A.shape[0], n = A.shape[1], mn = min(m,n)
	cdef np.ndarray[np.complex128_t,ndim=2] U = np.zeros((m,mn),dtype=np.complex128)
	cdef np.ndarray[np.double_t,ndim=1]     S = np.zeros((mn,) ,dtype=np.double)
	cdef np.ndarray[np.complex128_t,ndim=2] V = np.zeros((n,mn),dtype=np.complex128)
	# Compute SVD using TSQR algorithm
	retval = c_ztsqr_svd(&U[0,0],&S[0],&V[0,0],&A[0,0],m,n)
	if not retval == 0: raiseError('Problems computing TSQR SVD!')
	return U,S,V

@cr('math.tsqr_svd')
@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def tsqr_svd(real_full[:,:] A):
	'''
	Parallel Single value decomposition (SVD) using Lapack.
		U(m,n)   are the POD modes.
		S(n)     are the singular values.
		V(n,n)   are the right singular vectors.
	'''
	if real_full is np.complex128_t:
		return _ztsqr_svd(A)
	elif real_full is np.complex64_t:
		return _ctsqr_svd(A)
	elif real_full is double:
		return _dtsqr_svd(A)
	else:
		return _stsqr_svd(A)

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def _srandomized_qr(float[:,:] A, int r, int q, int seed):
	'''
	Parallel Randomized QR factorization using Lapack.
		Q(m,r)   
		B(r,n)  
	'''
	cdef int retval
	cdef int m = A.shape[0], n = A.shape[1]
	cdef np.ndarray[np.float32_t,ndim=2] Q = np.zeros((m,r),dtype=np.float32)
	cdef np.ndarray[np.float32_t,ndim=2] B = np.zeros((r,n),dtype=np.float32)
	# Compute SVD using randomized algorithm
	retval = c_srandomized_qr(&Q[0,0],&B[0,0],&A[0,0],m,n,r,q,seed)
	if not retval == 0: raiseError('Problems computing Randomized SVD!')
	return Q,B

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def _drandomized_qr(double[:,:] A, int r, int q, int seed):
	'''
	Parallel Randomized QR factorization using Lapack.
		Q(m,r)   
		B(n,n)    
	'''
	cdef int retval
	cdef int m = A.shape[0], n = A.shape[1]
	cdef np.ndarray[np.double_t,ndim=2] Q = np.zeros((m,r),dtype=np.double)
	cdef np.ndarray[np.double_t,ndim=2] B = np.zeros((r,n),dtype=np.double)
	# Compute SVD using randomized algorithm
	retval = c_drandomized_qr(&Q[0,0],&B[0,0],&A[0,0],m,n,r,q,seed)
	if not retval == 0: raiseError('Problems computing Randomized SVD!')
	return Q,B

@cr('math.randomized_qr')
@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def randomized_qr(real[:,:] A, const int r, const int q, const int seed=-1):
	'''
	Parallel Single value decomposition (SVD) using Lapack.
		Q(m,r)   
		B(n,r)   
	'''
	seed = <int>time(NULL) if seed < 0 else seed
	if real is double:
		return _drandomized_qr(A,r,q,seed)
	else:
		return _srandomized_qr(A,r,q,seed)

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def _sinit_qr_streaming(float[:,:] A, int r, int q, int seed):
	'''
	Parallel Randomized QR factorization using Lapack.
		Q(m,r)   
		B(r,n)  
	'''
	cdef int retval
	cdef int m = A.shape[0], n = A.shape[1]
	cdef np.ndarray[np.float32_t,ndim=2] Q = np.zeros((m,r),dtype=np.float32)
	cdef np.ndarray[np.float32_t,ndim=2] Y = np.zeros((m,r),dtype=np.float32)
	cdef np.ndarray[np.float32_t,ndim=2] B = np.zeros((r,n),dtype=np.float32)
	# Compute SVD using randomized algorithm
	retval = c_sinit_randomized_qr(&Q[0,0],&B[0,0],&Y[0,0],&A[0,0],m,n,r,q,seed)
	if not retval == 0: raiseError('Problems computing Randomized SVD!')
	return Q,B,Y

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def _dinit_qr_streaming(double[:,:] A, int r, int q, int seed):
	'''
	Parallel Randomized QR factorization using Lapack.
		Q(m,r)   
		B(n,n)    
	'''
	cdef int retval
	cdef int m = A.shape[0], n = A.shape[1]
	cdef np.ndarray[np.double_t,ndim=2] Q = np.zeros((m,r),dtype=np.double)
	cdef np.ndarray[np.double_t,ndim=2] Y = np.zeros((m,r),dtype=np.double)
	cdef np.ndarray[np.double_t,ndim=2] B = np.zeros((r,n),dtype=np.double)
	# Compute SVD using randomized algorithm
	retval = c_dinit_randomized_qr(&Q[0,0],&B[0,0],&Y[0,0],&A[0,0],m,n,r,q,seed)
	if not retval == 0: raiseError('Problems computing Randomized SVD!')
	return Q,B,Y

@cr('math.init_qr_streaming')
@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def init_qr_streaming(real[:,:] A, const int r, const int q, seed=None):
	'''
	Parallel Single value decomposition (SVD) using Lapack.
		Q(m,r)   
		B(n,r)   
	'''
	cdef unsigned int seed2 = <int>time(NULL) if seed == None else int(seed)
	if real is double:
		return _dinit_qr_streaming(A,r,q,seed2)
	else:
		return _sinit_qr_streaming(A,r,q,seed2)

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def _supdate_qr_streaming(float[:,:] Q1, float[:,:] B1, float[:,:] Yo, float[:,:] A, int r, int q, int seed):
	'''
	Parallel Randomized QR factorization using Lapack.
		Q(m,r)   
		B(r,n)  
	'''
	cdef int retval
	cdef int m  = A.shape[0], n = A.shape[1], n1 = B1.shape[1]
	cdef int n2 = n1+n
	cdef np.ndarray[np.float32_t,ndim=2] Q2 = np.zeros((m,r), dtype=np.float32)
	cdef np.ndarray[np.float32_t,ndim=2] B2 = np.zeros((r,n2),dtype=np.float32)
	cdef np.ndarray[np.float32_t,ndim=2] Yn = np.zeros((m,r), dtype=np.float32)
	# Compute SVD using randomized algorithm
	retval = c_supdate_randomized_qr(&Q2[0,0],&B2[0,0],&Yn[0,0],&Q1[0,0],&B1[0,0],&Yo[0,0],&A[0,0],m,n,n1,n2,r,q,seed)
	if not retval == 0: raiseError('Problems updating randomized QR!')
	return Q2,B2,Yn

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def _dupdate_qr_streaming(double[:,:] Q1, double[:,:] B1, double[:,:] Yo, double[:,:] A, int r, int q, int seed):
	'''
	Parallel Randomized QR factorization using Lapack.
		Q(m,r)   
		B(r,n)  
	'''
	cdef int retval
	cdef int m  = A.shape[0], n = A.shape[1], n1 = B1.shape[1]
	cdef int n2 = n1+n
	cdef np.ndarray[np.double_t,ndim=2] Q2 = np.zeros((m,r), dtype=np.double)
	cdef np.ndarray[np.double_t,ndim=2] B2 = np.zeros((r,n2),dtype=np.double)
	cdef np.ndarray[np.double_t,ndim=2] Yn = np.zeros((m,r), dtype=np.double)
	# Compute SVD using randomized algorithm
	retval = c_dupdate_randomized_qr(&Q2[0,0],&B2[0,0],&Yn[0,0],&Q1[0,0],&B1[0,0],&Yo[0,0],&A[0,0],m,n,n1,n2,r,q,seed)
	if not retval == 0: raiseError('Problems updating randomized QR!')
	return Q2,B2,Yn

@cr('math.qr_iteration')
@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def update_qr_streaming(real[:,:] A, real [:,:] Q1, real[:,:] B1, real[:,:] Yo, const int r, const int q):
	'''
	Parallel Single value decomposition (SVD) using Lapack.
		Q(m,r)   
		B(n,r)   
	'''
	cdef unsigned int seed = <int>time(NULL)
	if real is double:
		return _dupdate_qr_streaming(Q1,B1,Yo,A,r,q,seed)
	else:
		return _supdate_qr_streaming(Q1,B1,Yo,A,r,q,seed)

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def _srandomized_svd(float[:,:] A, int r, int q, int seed):
	'''
	Parallel Randomized Single value decomposition (SVD) using Lapack.
		U(m,n)   are the POD modes.
		S(n)     are the singular values.
		V(n,n)   are the right singular vectors.
	'''
	cdef int retval
	cdef int m = A.shape[0], n = A.shape[1], mn = min(m,n)
	cdef np.ndarray[np.float32_t,ndim=2] U = np.zeros((m,r),dtype=np.float32)
	cdef np.ndarray[np.float32_t,ndim=1] S = np.zeros((r,) ,dtype=np.float32)
	cdef np.ndarray[np.float32_t,ndim=2] V = np.zeros((r,n),dtype=np.float32)
	# Compute SVD using randomized algorithm
	retval = c_srandomized_svd(&U[0,0],&S[0],&V[0,0],&A[0,0],m,n,r,q,seed)
	if not retval == 0: raiseError('Problems computing Randomized SVD!')
	return U,S,V

@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def _drandomized_svd(double[:,:] A, int r, int q, int seed):
	'''
	Parallel Randomized Single value decomposition (SVD) using Lapack.
		U(m,n)   are the POD modes.
		S(n)     are the singular values.
		V(n,n)   are the right singular vectors.
	'''
	cdef int retval
	cdef int m = A.shape[0], n = A.shape[1], mn = min(m,n)
	cdef np.ndarray[np.double_t,ndim=2] U = np.zeros((m,r),dtype=np.double)
	cdef np.ndarray[np.double_t,ndim=1] S = np.zeros((r,) ,dtype=np.double)
	cdef np.ndarray[np.double_t,ndim=2] V = np.zeros((r,n),dtype=np.double)
	# Compute SVD using randomized algorithm
	retval = c_drandomized_svd(&U[0,0],&S[0],&V[0,0],&A[0,0],m,n,r,q,seed)
	if not retval == 0: raiseError('Problems computing Randomized SVD!')
	return U,S,V

@cr('math.randomized_svd')
@cython.initializedcheck(False)
@cython.boundscheck(False) # turn off bounds-checking for entire function
@cython.wraparound(False)  # turn off negative index wrapping for entire function
@cython.nonecheck(False)
@cython.cdivision(True)    # turn off zero division check
def randomized_svd(real[:,:] A, const int r, const int q, const int seed=-1):
	'''
	Parallel Single value decomposition (SVD) using Lapack.
		U(m,n)   are the POD modes.
		S(n)     are the singular values.
		V(n,n)   are the right singular vectors.
	'''
	seed = <int>time(NULL) if seed < 0 else seed
	if real is double:
		return _drandomized_svd(A,r,q,seed)
	else:
		return _srandomized_svd(A,r,q,seed)


from abc import ABC, abstractmethod
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import torch
import torch.utils
from torch.utils.data import DataLoader

from ... import pprint, cr # pyLOM/__init__.py
from ..utils import Dataset


class PINN(ABC):
    """
    This class represents a Physics-Informed Neural Network (PINN) model. It is an abstract class that needs to be subclassed to implement the pde_loss method.
    That method should compute the residual from the partial differential equation (PDE) and then compute the loss from it (usually by squaring the residual).

    Args:
        neural_net (torch.nn.Module): A neural network model that implements torch.nn.Module.
        device (str): The device to run the model on (e.g., 'cpu', 'cuda').

    Attributes:
        device (str): The device the model is running on.
        model (torch.nn.Module): The neural network model.

    """

    def __init__(self, neural_net, device):
        self.device = device
        self.model = neural_net.to(device)

    def __call__(self, x):
        """
        Forward pass of the PINN model.

        Args:
            x (torch.Tensor): The input tensor with the PDE input parameters.

        Returns:
            torch.Tensor: The output tensor, i.e. the solution for the PDE on x.

        """
        return self.model(x)

    def _prepare_input_variables(self, x_batch):
        """
        Prepares the input variables for training.

        Args:
            x_batch (torch.Tensor): The input batch tensor.

        Returns:
            List[torch.Tensor]: The list of prepared input variables.

        """
        input_variables = []
        
        for input_variable in range(x_batch.shape[1]):
            flow_variable = x_batch[:, input_variable : input_variable + 1]
            flow_variable.requires_grad_(True)
            input_variables.append(flow_variable)
        return input_variables

    def _get_dataloader(self, dataset, batch_size=None):
        """
        Creates data loaders for training.

        Args:
            x (torch.Tensor): The input tensor.
            y (torch.Tensor, optional): The target tensor. Defaults to None.
            batch_size (int, optional): The batch size. Defaults to None.

        Returns:
            Union[torch.utils.data.DataLoader, List[torch.Tensor]]: The data loaders.

        """
        if batch_size is not None:
            data_loader = DataLoader(dataset, batch_size=batch_size, shuffle=False, num_workers=1, persistent_workers=True)
        else:
            data_loader = [dataset[:]]
        return data_loader

    def bc_data_loss(self, pred, y, boundary_conditions, use_bfloat16=False):
        """
        Computes the loss from boundary conditions and data.

        Args:
            pred (torch.Tensor): The predicted output tensor.
            y (torch.Tensor): The target tensor.
            boundary_conditions (List[BoundaryCondition]): The list of boundary conditions.
            use_bfloat16 (bool, optional): Whether to use bfloat16 precision. Defaults to False.

        Returns:
            List[torch.Tensor]: The list of loss tensors.

        """
        if use_bfloat16:
            with torch.autocast("cuda", dtype=torch.bfloat16):
                bc_losses = [bc.loss(self.model(bc.points.to(self.device))) for bc in boundary_conditions]
        else:
            bc_losses = [bc.loss(self.model(bc.points.to(self.device))) for bc in boundary_conditions]

        if y is not None:
            data_loss = torch.nn.functional.mse_loss(pred, y.to(self.device))
            bc_losses.append(data_loss)
        return bc_losses

    def compute_loss(self, x, y, boundary_conditions, use_bfloat16=False):
        """
        Computes the total loss for training.

        Args:
            x (torch.Tensor): The input tensor.
            y (torch.Tensor): The target tensor.
            boundary_conditions (List[BoundaryCondition]): The list of boundary conditions.
            use_bfloat16 (bool, optional): Whether to use bfloat16 precision. Defaults to False.

        Returns:
            List[torch.Tensor]: The list of loss tensors.

        """
        input_variables = self._prepare_input_variables(x)
        input_tensor = torch.cat(input_variables, dim=1).to(self.device)

        if use_bfloat16:
            with torch.autocast("cuda", dtype=torch.bfloat16):
                pred = self.model(input_tensor)
        else:
            pred = self.model(input_tensor)

        return [self.pde_loss(pred, *input_variables)] + self.bc_data_loss(pred, y, boundary_conditions, use_bfloat16)

    @cr("PINN.fit")
    def fit(
        self,
        train_dataset: Dataset,
        optimizer_class=torch.optim.Adam,
        optimizer_params={},
        lr_scheduler_class=None,
        lr_scheduler_params={},
        epochs=1000,
        boundary_conditions=[],
        update_logs_steps=1,
        loaded_logs=None,
        batch_size=None,
        eval_dataset: Dataset = None,
        use_bfloat16=False,
        **kwargs,
    ):
        """
        Trains the PINN model.

        Args:
            train_dataset (Dataset): The training dataset. If the dataset returns a tuple, the first element is the input and the second element is the target. If not, the PINN is trained without simulation data.
            optimizer_class (torch.optim.Optimizer, optional): The optimizer class. Defaults to ``torch.optim.Adam``.
            optimizer_params (dict, optional): The optimizer parameters. Defaults to ``{}``.
            lr_scheduler_class (torch.optim.lr_scheduler._LRScheduler, optional): The learning rate scheduler class. Defaults to ``None``.
            lr_scheduler_params (dict, optional): The learning rate scheduler parameters. Defaults to ``{}``.
            epochs (int, optional): The number of epochs to train for. Defaults to ``1000``.
            boundary_conditions (List[BoundaryCondition], optional): The list of boundary conditions. Defaults to ``[]``.
            update_logs_steps (int, optional): The interval for updating the progress. Defaults to ``100``.
            loaded_logs (dict, optional): Loaded training logs to be used as initial logs. Defaults to ``None``.
            batch_size (int, optional): The batch size. If none, the batch size will be equal to the number of collocation points given on `train_dataset`. Defaults to ``None``.
            eval_dataset (BaseDataset, optional): The evaluation dataset. Defaults to ``None``.
            use_bfloat16 (bool, optional): Whether to use bfloat16 precision. Defaults to ``False``.
            **kwargs: Additional keyword arguments.

        Returns:
            dict: The training logs.

        """
        logs = (
            loaded_logs
            if loaded_logs is not None
            else {
                "loss_from_pde": [],
                "loss_from_data_and_bc": [],
                "total_loss": [],
            }
        )

        train_data_loader = self._get_dataloader(train_dataset, batch_size)
        test_data_loader = None
        if eval_dataset is not None:
            test_data_loader = self._get_dataloader(eval_dataset, batch_size)
            if "test_loss" not in logs:
                logs["test_loss"] = []

        optimizer = optimizer_class(self.model.parameters(), **optimizer_params)
        if lr_scheduler_class is not None:
            lr_scheduler = lr_scheduler_class(optimizer, **lr_scheduler_params)

        def closure():
            x_batch = batch[0].to(self.device)
            y_batch = batch[1].to(self.device) if len(batch) == 2 else None

            optimizer.zero_grad()
            losses = self.compute_loss(x_batch, y_batch, boundary_conditions, use_bfloat16)
            loss = sum(losses)
            loss.backward()
            
            loss_from_pde = losses[0].item()
            logs["loss_from_pde"].append(loss_from_pde)
            logs["loss_from_data_and_bc"].append(loss.item() - loss_from_pde)
            logs["total_loss"].append(loss.item())

            if update_logs_steps != 0 and (epoch % update_logs_steps == 0):
                extended_desc = ''
                if len(losses) > 1:
                    extended_desc = f", data/bc losses: [{', '.join(f'{x:.4e}' for x in losses[1:])}]"
                if 'test_loss' in logs and len(logs['test_loss']) > 0:
                    extended_desc += f", test loss: {logs['test_loss'][-1]:.4e}"
                desc = f"Epoch {epoch+1}/{epochs} Iteration {closure.iteration}. Pde loss: {loss_from_pde:.4e}" + extended_desc
                pprint(0, desc)

            closure.iteration += 1
            return loss

        self.model.train()

        for epoch in range(epochs):
            closure.iteration = 0

            for batch in train_data_loader: 
                optimizer.step(closure=closure)
                if lr_scheduler_class is not None:
                    lr_scheduler.step()

            if test_data_loader is not None:
                self.model.eval()
                test_loss = 0
                for batch in test_data_loader:
                    x_batch = batch[0].to(self.device)
                    y_batch = batch[1].to(self.device) if len(batch) == 2 else None
                    losses = self.compute_loss(x_batch, y_batch, boundary_conditions)
                    test_loss += sum(losses).item()
                logs["test_loss"].append(test_loss / len(test_data_loader))
                self.model.train()

        return logs
    
    @cr("PINN.predict")
    def predict(self, X: Dataset, **kwargs) -> np.ndarray:
        """
        Predicts for the input dataset.

        Args:
            X (Dataset): The input dataset.

        Returns:
            np.ndarray: The predictions of the model.

        """
        self.model.eval()
        data = X[:]
        input_data = data[0] # keep only the input data
        return self.model(input_data.to(self.device)).detach().cpu().numpy()

    def __repr__(self):
        """
        Returns a string representation of the PINN model.

        Returns:
            str: The string representation.

        """
        pprint(0, f"Number of parameters: {sum(p.numel() for p in self.model.parameters())}")
        return self.model.__repr__()

    def plot_training_logs(self, logs):
        """
        Plots the training logs.

        Args:
            logs (dict): The training logs.

        """
        plt.figure(figsize=(10, 6))
        plt.plot(logs["loss_from_pde"], label="PDE Loss")
        plt.plot(logs["loss_from_data_and_bc"], label="Data Conditions and BC Loss")
        plt.plot(logs["total_loss"], label="Total Loss")
        if "test_loss" in logs and len(logs["test_loss"]) != 0:
            total_epochs = len(logs["test_loss"]) 
            total_iters = len(logs["total_loss"]) 
            iters_per_epoch = total_iters // total_epochs
            plt.plot(np.arange(iters_per_epoch + total_iters % total_epochs, total_iters+1, step=iters_per_epoch), logs["test_loss"], label="Test Loss")
        plt.xlabel("Iteration")
        plt.ylabel("Loss")
        plt.yscale("log")
        plt.title("Training Losses")
        plt.legend()
        plt.show()

    @abstractmethod
    def pde_loss(self, pred, *input_variables):
        """
        Computes the loss from the partial differential equation (PDE).

        Args:
            pred (torch.Tensor): The predicted output tensor.
            *input_variables (torch.Tensor): The input variables for the PDE. e.g. x, y, t.

        Returns:
            torch.Tensor: The loss tensor.

        """
        pass

    def save(self, path):
        """
        Saves the model to a file using torchscript.

        Args:
            path (str): The path to save the model.

        """

        path = Path(path)
        scripted_model = torch.jit.script(self.model)
        scripted_model.save(path)
        

    @classmethod
    def load(cls, path, device='cpu'):
        """
        Loads the model from a file.

        Args:
            path (str): The path to load the model.
            neural_net (torch.nn.Module): The neural network model.
            device (str, optional): The device to run the model on. Defaults to 'cpu'.

        Returns:
            PINN: The loaded PINN model.

        """
        model = torch.jit.load(path, map_location=device)
        return cls(neural_net=model, device=device)

class BoundaryCondition(ABC):
    """
    Abstract base class for defining boundary conditions. You need to implement the `loss` method to use a custom boundary condition.

    Args:
        points (Tensor): The points where the boundary condition is defined.

    Attributes:
        points (Tensor): The points where the boundary condition is defined.
    """
    def __init__(self, points):
        self._points = points
        self._points.requires_grad_(True)
    
    @abstractmethod
    def loss(self, pred):
        """
        Computes the loss for the given prediction.

        Args:
            pred (Tensor): The predicted values on the points where the boundary condition is defined.

        Returns:
            Tensor: The loss value.
        """
        pass

    @property
    def points(self):
        return self._points


class DirichletCondition(BoundaryCondition):
    """
    This class represents a Dirichlet boundary condition.

    Args:
        points (Tensor): The predicted values on the points where the boundary condition is defined.
        values (Tensor): The values of the boundary condition.
    """
    def __init__(self, points, values):
        super().__init__(points)
        self._values = values

    def loss(self, pred):
        return torch.mean((pred - self._values) ** 2)

class BurgersPINN(PINN):
    r"""
    This class represents a Physics-Informed Neural Network (PINN) model for the Burgers' equation.
    The model predictions have 1 column, the velocity field :math:`u`.

    .. math::
        \frac{\partial u}{\partial t} + u\frac{\partial u}{\partial x} = \nu\frac{\partial^2u}{\partial x^2}

    Args:
        neural_net (torch.nn.Module): The neural network model.
        device (str): The device to run the model on (e.g., 'cpu', 'cuda').
        viscosity (float): The viscosity coefficient.
    """
    def __init__(self, neural_net, device, viscosity=0.01):
        super().__init__(neural_net, device)
        self.viscosity = viscosity

    def pde_loss(self, pred, *input_variables):
        t, x = input_variables
        u = pred
        u_t = torch.autograd.grad(u, t, grad_outputs=torch.ones_like(u), create_graph=True)[0]
        u_x = torch.autograd.grad(u, x, grad_outputs=torch.ones_like(u), create_graph=True)[0]
        u_xx = torch.autograd.grad(u_x, x, grad_outputs=torch.ones_like(u_x), create_graph=True)[0]

        f = u_t + u * u_x - (self.viscosity / torch.pi) * u_xx

        return (f ** 2).mean()

class NavierStokesIncompressible2DPINN(PINN):
    r"""
    This class represents a Physics-Informed Neural Network (PINN) model for the incompressible steady 2D Navier-Stokes equations.
    The model predictions have 3 columns, the velocity field :math:`(u, v)` and the pressure :math:`p` fields.
    
    .. math::
        \begin{align*}
        u \frac{\partial u}{\partial x} + v \frac{\partial u}{\partial y} + \frac{\partial p}{\partial x} - \frac{1}{Re} \left( \frac{\partial^2 u}{\partial x^2} + \frac{\partial^2 u}{\partial y^2} \right) &= 0, \\
        u \frac{\partial v}{\partial x} + v \frac{\partial v}{\partial y} + \frac{\partial p}{\partial y} - \frac{1}{Re} \left( \frac{\partial^2 v}{\partial x^2} + \frac{\partial^2 v}{\partial y^2} \right) &= 0, \\
        \frac{\partial u}{\partial x} + \frac{\partial v}{\partial y} &= 0.
        \end{align*}

    Args:
        neural_net (torch.nn.Module): The neural network model.
        device (str): The device to run the model on (e.g., 'cpu', 'cuda').
        Re (float): The Reynolds number.
    """

    def __init__(self, neural_net, device, Re=50):
        super().__init__(neural_net, device)
        self.Re = Re
    
    def pde_loss(self, pred , x, y):

        u, v, p= pred[:, 0:1], pred[:, 1:2], pred[:, 2:3]        

        # Spatial gradients
        u_x = torch.autograd.grad(u, x, grad_outputs=torch.ones_like(u), create_graph=True)[0]
        u_y = torch.autograd.grad(u, y, grad_outputs=torch.ones_like(u), create_graph=True)[0]
        
        v_x = torch.autograd.grad(v, x, grad_outputs=torch.ones_like(v), create_graph=True)[0]
        v_y = torch.autograd.grad(v, y, grad_outputs=torch.ones_like(v), create_graph=True)[0]
        
        p_x = torch.autograd.grad(p, x, grad_outputs=torch.ones_like(p), create_graph=True)[0]
        p_y = torch.autograd.grad(p, y, grad_outputs=torch.ones_like(p), create_graph=True)[0]
        
        # Second order spatial gradients
        u_xx = torch.autograd.grad(u_x, x, grad_outputs=torch.ones_like(u_x), create_graph=True)[0]
        u_yy = torch.autograd.grad(u_y, y, grad_outputs=torch.ones_like(u_y), create_graph=True)[0]
        
        v_xx = torch.autograd.grad(v_x, x, grad_outputs=torch.ones_like(v_x), create_graph=True)[0]
        v_yy = torch.autograd.grad(v_y, y, grad_outputs=torch.ones_like(v_y), create_graph=True)[0]
        
        # Continuity equation
        continuity = u_x + v_y
        
        # Momentum equations
        momentum_u = u * u_x + v * u_y + p_x - (1 / self.Re) * (u_xx + u_yy)
        momentum_v = u * v_x + v * v_y + p_y - (1 / self.Re) * (v_xx + v_yy)

        # Total loss
        loss_continuity = torch.mean(continuity ** 2)
        loss_momentum_u = torch.mean(momentum_u ** 2)
        loss_momentum_v = torch.mean(momentum_v ** 2)
        
        loss = loss_continuity + loss_momentum_u + loss_momentum_v
        return loss

class Euler2DPINN(PINN):
    # TODO: review the equations once sphinx documentation is working
    r"""
    This class represents a Physics-Informed Neural Network (PINN) model for the 2D Euler equations.
    The model predictions have 4 columns, the density :math:`\rho`, the velocity field :math:`(u, v)` and the total energy :math:`E` fields.

    .. math::
        \begin{align*}
        \frac{\partial \rho}{\partial t} + \frac{\partial (\rho u)}{\partial x} + \frac{\partial (\rho v)}{\partial y} &= 0, \\
        \frac{\partial (\rho u)}{\partial t} + \frac{\partial (\rho u^2 + p)}{\partial x} + \frac{\partial (\rho uv)}{\partial y} &= 0, \\
        \frac{\partial (\rho v)}{\partial t} + \frac{\partial (\rho uv)}{\partial x} + \frac{\partial (\rho v^2 + p)}{\partial y} &= 0, \\
        \frac{\partial (\rho E)}{\partial t} + \frac{\partial (u(\rho E + p))}{\partial x} + \frac{\partial (v(\rho E + p))}{\partial y} &= 0.
        \end{align*}

    Args:
        neural_net (torch.nn.Module): The neural network model.
        device (str): The device to run the model on (e.g., 'cpu', 'cuda').
    """

    GAMMA = 1.4

    def pde_loss(self, pred, *input_variables):

        x_coord = input_variables[0]
        y_coord = input_variables[1]
        rho, u, v, E = pred[:, 0:1], pred[:, 1:2], pred[:, 2:3], pred[:, 3:4]

        p = (Euler2DPINN.GAMMA - 1.0) * (E - 0.5 * rho * (u**2 + v**2))

        # F
        F1 = rho * u
        F2 = rho * u**2 + p
        F3 = rho * u * v
        F4 = u * (E + p)

        # G
        G1 = rho * v
        G2 = rho * u * v
        G3 = rho * v**2 + p
        G4 = v * (E + p)

        dF1_dx = torch.autograd.grad(F1, x_coord, grad_outputs=torch.ones_like(F1), create_graph=True)[0]
        dF2_dx = torch.autograd.grad(F2, x_coord, grad_outputs=torch.ones_like(F2), create_graph=True)[0]
        dF3_dx = torch.autograd.grad(F3, x_coord, grad_outputs=torch.ones_like(F3), create_graph=True)[0]
        dF4_dx = torch.autograd.grad(F4, x_coord, grad_outputs=torch.ones_like(F4), create_graph=True)[0]

        dG1_dy = torch.autograd.grad(G1, y_coord, grad_outputs=torch.ones_like(G1), create_graph=True)[0]
        dG2_dy = torch.autograd.grad(G2, y_coord, grad_outputs=torch.ones_like(G2), create_graph=True)[0]
        dG3_dy = torch.autograd.grad(G3, y_coord, grad_outputs=torch.ones_like(G3), create_graph=True)[0]
        dG4_dy = torch.autograd.grad(G4, y_coord, grad_outputs=torch.ones_like(G4), create_graph=True)[0]

        residual_mass_conserv = dF1_dx + dG1_dy
        residual_momentum_x = dF2_dx + dG2_dy
        residual_momentum_y = dF3_dx + dG3_dy
        residual_energy = dF4_dx + dG4_dy

        loss_mass_conserv = torch.mean((residual_mass_conserv) ** 2) 
        loss_momentum_x = torch.mean((residual_momentum_x) ** 2)  
        loss_momentum_y = torch.mean((residual_momentum_y) ** 2)  
        loss_energy = torch.mean((residual_energy) ** 2) 

        loss = loss_mass_conserv + loss_momentum_x + loss_momentum_y + loss_energy
        return loss


#!/usr/bin/env python
#
# pyLOM - Python Low Order Modeling.
#
# Utils - Error handling routines.
#
# Last rev: 09/07/2021
from __future__ import print_function, division

import sys, numpy as np
from .mpi import MPI_RANK, MPI_COMM


def round(value,precision):
	'''
	Truncate array by a certain precision
	'''
	fact  = 10**precision
	return np.round(value*fact)/fact


def raiseError(errmsg):
	'''
	Raise a controlled error and abort execution on
	all processes.
	'''
	print('%d - %s' % (MPI_RANK,errmsg),file=sys.stderr,flush=True)
	MPI_COMM.Abort(1)


def raiseWarning(warnmsg,allranks=False):
	'''
	Raise a controlled warning but don't abort execution on
	all processes.
	'''
	if allranks:
		print('Warning! %d - %s' % (MPI_RANK,warnmsg),file=sys.stderr,flush=True)
	else:
		if MPI_RANK == 0: print(0,'Warning! %s' % (warnmsg),file=sys.stderr,flush=True)


#!/usr/bin/env python
#
# pyLOM - Python Low Order Modeling.
#
# NN optimizer routines using optuna.
#
# Last rev: 02/10/2024

import json

from typing         import Callable, Dict
from ..utils.errors import raiseError
from ..             import pprint

# Add optuna as an optional dependency
try:
    import optuna
    from optuna.trial      import TrialState
    from optuna.exceptions import TrialPruned


    class OptunaOptimizer():
        """
        Args:
            optimization_params (Dict): A dictionary containing the parameters to optimize.
            n_trials (int): The number of trials to run. Default is ``100``.
            direction (str): The direction to optimize. Can be 'minimize' or 'maximize'. Default is ``'minimize'``.
            pruner (optuna.pruners.BasePruner): The pruner to use. Default is ``None``.
            save_dir (str): The directory to save the best parameters. Default is ``None``.
        """
        def __init__(
            self,
            optimization_params: Dict,
            n_trials: int = 100,
            direction: str = 'minimize',
            pruner: optuna.pruners.BasePruner = None,
            save_dir: str = None
        ):
            self.num_trials = n_trials
            self.direction = direction
            self.pruner = pruner
            self._optimization_params = optimization_params
            self.save_dir = save_dir

        @property
        def optimization_params(self) -> Dict:
            """
            Get the optimization parameters.
            """
            return self._optimization_params

        def optimize(
                self, 
                objective_function: Callable[[optuna.Trial], float],
        ) -> Dict:
            """
            Optimize a model given an objective function.
            
            Args:
                objective_function (Callable): The objective function to optimize. The function should take a `optuna.Trial` object as input and return a float.
                
            Returns:
                Dict: The best parameters obtained from the optimization.
            """
            study = optuna.create_study(direction=self.direction, pruner=self.pruner)
            study.optimize(objective_function, n_trials=self.num_trials)
            if self.save_dir is not None:
                file_path = str(self.save_dir) + '/best_params.json'
                json.dump(study.best_params, open(file_path, 'w'))

            self._print_optimization_report(study)
            
            return study.best_params
        
        def _print_optimization_report(self, study):
            pruned_trials = study.get_trials(deepcopy=False, states=[TrialState.PRUNED])
            completed_trials = study.get_trials(deepcopy=False, states=[TrialState.COMPLETE])
            
            # Optimization report
            pprint(0, "\nStudy statistics: ")
            pprint(0, "  Number of finished trials: ", len(study.trials))
            pprint(0, "  Number of pruned trials: ", len(pruned_trials))
            pprint(0, "  Number of completed trials: ", len(completed_trials))

            trial = study.best_trial
            pprint(0, "Best trial:")
            pprint(0, "  Value: ", trial.value)
            pprint(0, "  Params: ")
            for key, value in trial.params.items():
                pprint(0, "    {}: {}".format(key, value))
            pprint(0, "\n")

except:
    def TrialState():
        raiseError("Package optuna should be installed")
    
    def TrialPruned():
        raiseError("Package optuna should be installed")

    class OptunaOptimizer():
        """
        Args:
            optimization_params (Dict): A dictionary containing the parameters to optimize.
            n_trials (int): The number of trials to run. Default is ``100``.
            direction (str): The direction to optimize. Can be 'minimize' or 'maximize'. Default is ``'minimize'``.
            pruner (optuna.pruners.BasePruner): The pruner to use. Default is ``None``.
            save_dir (str): The directory to save the best parameters. Default is ``None``.
        """
        def __init__(self,*args,**kwargs):
            raiseError("Package optuna should be installed")


#!/usr/bin/env python
#
# pyLOM - Python Low Order Modeling.
#
# POD plotting utilities.
#
# Last rev: 27/10/2021
from __future__ import print_function, division

import numpy as np
import matplotlib.pyplot as plt

from ..vmmath      import fft
from ..utils       import gpu_to_cpu
from ..utils.plots import plotResidual, plotSnapshot


def plotMode(V,t,modes=np.array([1],np.int32),fftfun=fft,scale_freq=1.,fig=[],ax=[],cmap=None):
	'''
	Given U, VT and a mode, plot their
	representation in a figure.
	'''
	V = gpu_to_cpu(V)
	for imode, mode in enumerate(modes):
		if len(fig) < imode + 1:
			fig.append( plt.figure(figsize=(8,6),dpi=100) )
		if len(ax) < imode + 1:
			ax.append( fig[imode].subplots(2,1,gridspec_kw = {'hspace':0.5}) )
		fig[imode].suptitle('Mode %d'%(mode-1))
	   	# Plot the temporal evolution of the mode
		ax[imode][0].plot(t,V[mode-1,:],'b')
		ax[imode][0].set_title('Temporal mode')
		# Plot frequency representation of the mode
		if V.shape[1] % 2 == 0:
			freq,psd = fftfun(t,V[mode-1,:],equispaced=False)
		else:
			freq,psd = fftfun(t[:-1],V[mode-1,:-1],equispaced=False)
		freq *= scale_freq
		#L = int(np.floor(freq.shape[0]/2))
		ax[imode][1].plot(freq,psd, 'b')
		ax[imode][1].set_title('Power Spectrum')
		ax[imode][1].set_xlabel('St')
		ax[imode][1].set_xlim([0,1])
	return fig, ax


#!/usr/bin/env python
#
# Example of POD.
#
# Last revision: 19/07/2021
from __future__ import print_function, division

import mpi4py
mpi4py.rc.recv_mprobe = False

import numpy as np
import pyLOM

pyLOM.gpu_device(gpu_per_node=4) # Detect GPU configuration


## Parameters
DATAFILE = './DATA/jetLES.h5'
VARIABLE = 'PRESS'


## Data loadingx
m = pyLOM.Mesh.load(DATAFILE)
d = pyLOM.Dataset.load(DATAFILE,ptable=m.partition_table).to_gpu([VARIABLE]) # Send to GPU if available
X = d[VARIABLE][:,::10].copy()
t = d.get_variable('time')[::10].copy()


## Run POD
PSI,S,V = pyLOM.POD.run(X,remove_mean=False) # PSI are POD modes
if pyLOM.utils.is_rank_or_serial(root=0): pyLOM.POD.plotResidual(S)
# Truncate according to a residual
PSI,S,V = pyLOM.POD.truncate(PSI,S,V,r=5e-6)
pyLOM.POD.save('results.h5',PSI,S,V,d.partition_table,nvars=1,pointData=d.point)
# Reconstruct the flow
X_POD = pyLOM.POD.reconstruct(PSI,S,V)
# Compute RMSE
rmse = pyLOM.math.RMSE(X_POD,X)
pyLOM.pprint(0,'RMSE = %e'%rmse)


## Dump to ParaView
# Spatial modes
d.add_field('spatial_modes_P',6,pyLOM.POD.extract_modes(PSI,1,len(d),modes=[1,4,6,2,5,3]))
pyLOM.io.pv_writer(m,d.to_cpu(['spatial_modes_P']),'modes',basedir='out/modes',instants=[0],times=[0.],vars=['spatial_modes_P'],fmt='vtkh5')
pyLOM.POD.plotSnapshot(m,d.to_cpu(['spatial_modes_P']),vars=['spatial_modes_P'],instant=0,component=0,cmap='jet',cpos='xy')

# Temporal evolution
d.add_field('PRESR',1,X_POD)
pyLOM.io.pv_writer(m,d.to_cpu(['PRESS','PRESR']),'flow',basedir='out/flow',instants=np.arange(t.shape[0],dtype=np.int32),times=t,vars=['PRESS','PRESR'],fmt='vtkh5')
pyLOM.POD.plotSnapshot(m,d.to_cpu(['PRESS','PRESR']),vars=['PRESR'],instant=0,component=0,cmap='jet',cpos='xy')


## Plot POD mode
if pyLOM.utils.is_rank_or_serial(0):
	# 0 - module, 1,2 - components
	pyLOM.POD.plotMode(V,t,modes=[1,2,3,4],scale_freq=2.56)


## Show and print timings
pyLOM.cr_info()
pyLOM.show_plots()


#!/usr/bin/env cpython
#
# pyLOM - Python Low Order Modeling.
#
# Math operations module - geometry.
#
# Last rev: 27/10/2021
from __future__ import print_function, division

import numpy as np

from ..utils.gpu import cp
from ..utils     import cr_nvtx as cr, mpi_reduce


@cr('math.cellCenters')
def cellCenters(xyz:np.ndarray,conec:np.ndarray) -> np.ndarray:
	r'''
	Compute the cell centers given a list 
	of elements.

	Args:
		xyz (np.ndarray):   node positions
		conec (np.ndarray): connectivity array

	Returns:
		np.ndarray: center positions
	'''
	p = cp if type(xyz) is cp.ndarray else np
	xyz_cen = p.zeros((conec.shape[0],xyz.shape[1]),xyz.dtype)
	for ielem in range(conec.shape[0]):
		# Get the values of the field and the positions of the element
		c = conec[ielem,conec[ielem,:]>=0]
		xyz_cen[ielem,:] = p.mean(xyz[c,:],axis=0)
	return xyz_cen

@cr('math.normals')
def normals(xyz:np.ndarray,conec:np.ndarray) -> np.ndarray:
	r'''
	Compute the cell normals given a list 
	of elements.

	Args:
		xyz (np.ndarray):   node positions
		conec (np.ndarray): connectivity array

	Returns:
		np.ndarray: cell normals
	'''
	p = cp if type(xyz) is cp.ndarray else np
	normals = p.zeros(((conec.shape[0],3)),xyz.dtype)
	for ielem in range(conec.shape[0]):
		# Get the values of the field and the positions of the element
		c     = conec[ielem,conec[ielem,:]>=0]
		xyzel =  xyz[c,:]
		# Compute centroid
		cen  = p.mean(xyzel,axis=0)
		# Compute normal
		for inod in range(len(c)):
			u = xyzel[inod]   - cen
			v = xyzel[inod-1] - cen
			normals[ielem,:] += 0.5*p.cross(u,v)
	return normals

@cr('math.euclidean_d')
def euclidean_d(X:np.ndarray) -> np.ndarray:
	r'''
	Compute the Euclidean distances between simulations.

	Args:
		X (np.ndarray): NxM Data matrix with N points in the mesh for M simulations

	Returns:
		np.ndarray: MxM distance matrix 
	'''
	p = cp if type(X) is cp.ndarray else np
	# Extract dimensions
	_,M = X.shape
	# Initialize distance matrix
	D = p.zeros((M,M),X.dtype)
	for i in range(M):
		for j in range(i+1,M,1):
			# Local sum on the partition
			d2 = p.sum((X[:,i]-X[:,j])*(X[:,i]-X[:,j]))
			# Global sum over the partitions
			dG = p.sqrt(mpi_reduce(d2,all=True))
			# Fill output
			D[i,j] = dG
			D[j,i] = dG
	# Return the mdistance matrix
	return D



#!/usr/bin/env python
#
# pyLOM - Python Low Order Modeling.
#
# IO Module
#
# Last rev: 20/10/2021

# Pickle and HDF5 exchange format
from .io_pkl  import pkl_load, pkl_save
from .io_h5   import h5_load_dset, h5_save_dset, h5_append_dset, h5_save_mesh, h5_load_mesh, h5_save_POD, h5_load_POD, h5_save_DMD, h5_load_DMD, h5_save_SPOD, h5_load_SPOD, h5_save_VAE

# VTK HDF5 3D format
from .io_vtkh5 import vtkh5_save_mesh, vtkh5_link_mesh, vtkh5_save_field

# Output to paraview
from .io_paraview import pv_writer

del io_pkl, io_h5, io_paraview



#!/usr/bin/env cpython
#
# pyLOM - Python Low Order Modeling.
#
# Math operations module - vector/matrix math.
#
# Last rev: 27/10/2021
from __future__ import print_function, division

import numpy as np

from ..utils.gpu import cp
from ..utils     import cr_nvtx as cr, mpi_reduce, gpu_to_cpu


## Python functions
@cr('math.transpose')
def transpose(A:np.ndarray) -> np.ndarray:
	r'''
	Transposed of matrix A

	Args:
		A (np.ndarray): Matrix to be transposed
	
	Results
		np.ndarray: Transposed matrix
	'''
	p = cp if type(A) is cp.ndarray else np
	return p.transpose(A)

@cr('math.vector_norm')
def vector_sum(v:np.ndarray,start:int=0) -> float:
	r'''
	Sum of a vector

	Args:
		v (np.ndarray): a vector
		start (int): position of the vector where to start the sum
	
	Result:
		float: sum of the vector
	'''
	p = cp if type(v) is cp.ndarray else np
	return p.sum(v[start:])

@cr('math.vector_norm')
def vector_norm(v:np.ndarray,start:int=0) -> float:
	r'''
	L2 norm of a vector
	
	Args:
		v (np.ndarray): a vector
		start (int): position of the vector where to start the norm
	
	Result:
		float: norm of the vector
	'''
	p = cp if type(v) is cp.ndarray else np
	return p.linalg.norm(v[start:],2)

@cr('math.vector_mean')
def vector_mean(v:np.ndarray,start:int=0) -> float:
	r'''
	Mean of a vector

	Args:
		v (np.ndarray): a vector
		start (int): position of the vector where to start the mean
	
	Result:
		float: mean of the vector
	'''
	p = cp if type(v) is cp.ndarray else np
	return p.mean(v[start:])

@cr('math.matmul')
def matmul(A:np.ndarray,B:np.ndarray) -> np.ndarray:
	r'''
	Matrix multiplication 
	C = A x B

	Args:
		A (np.ndarray): Matrix A (M,Q)
		B (np.ndarray): Matrix B (Q,N)
	
	Result:
		np.ndarray: Resulting matrix C (M,N)
	'''
	p = cp if type(A) is cp.ndarray else np
	return p.matmul(A,B)

@cr('math.matmulp')
def matmulp(A:np.ndarray,B:np.ndarray) -> np.ndarray:
	r'''
	Matrix multiplication in parallel
	C = A x B 

	.. warning::
	A and B are distributed along the processors and C is the same for all of them

	Args:
		A (np.ndarray): Matrix A (M,Q)
		B (np.ndarray): Matrix B (Q,N)
	
	Result:
		np.ndarray: Resulting matrix C (M,N)
	'''
	p = cp if type(A) is cp.ndarray else np
	aux = p.matmul(A,B)
	return mpi_reduce(aux, root = 0, op = 'sum', all = True)

@cr('math.vecmat')
def vecmat(v:np.ndarray,A:np.ndarray) -> np.ndarray:
	r'''
	Vector times a matrix 
	C = v x A

	Args:
		v (np.ndarray): Vector v (M,)
		A (np.ndarray): Matrix A (M,N)
	
	Result:
		np.ndarray: Resulting matrix C (M,N)
	'''
	p = cp if type(v) is cp.ndarray else np
	C = p.zeros_like(A)
	for ii in range(v.shape[0]):
		C[ii,:] = v[ii]*A[ii,:]
	return C

@cr('math.argsort')
def argsort(v:np.ndarray) -> np.ndarray:
	r'''
	Returns the indices that sort a vector

	Args:
		v (np.ndarray): Vector v (M,)
	
	Result:
		np.ndarray: Indices that sort v (M,)
	'''
	p = cp if type(v) is cp.ndarray else np
	return p.argsort(v)

@cr('math.diag')
def diag(A:np.ndarray) -> np.ndarray:
	r'''
	If A is a matrix it returns its diagonal, if its a vector it returns
	a diagonal matrix with A in its diagonal

	Args:
		A (np.ndarray): Matrix A (M,N)
	
	Result:
		np.ndarray: Diagonal of A (M,)
	'''
	p = cp if type(A) is cp.ndarray else np
	return p.diag(A)

@cr('math.eigen')
def eigen(A:np.ndarray) -> np.ndarray:
	r'''
	Eigenvalues and eigenvectors.

	.. warning::
	GPU implementation of this algoritm is still slow and
	thus will be executed purely on CPU level until
	cupy implements linalg.eig

	Args:
		A (np.ndarray): Matrix A (M,N)
	
	Result:
		np.ndarray: the real eigenvalues, real(M)
		np.ndarray: the imaginary eigenvalues, imag(M)
		np.ndarray: the right eigenvectors, vecs(M,M)
	'''
	A = gpu_to_cpu(A)
	w,vecs = np.linalg.eig(A)
	real   = np.real(w)
	imag   = np.imag(w)
	return real, imag, vecs

@cr('math.polar')
def polar(real:np.ndarray, imag:np.ndarray) -> np.ndarray:
	r'''
	Present a complex number in its polar form given its real and imaginary part

	Args:
		real (np.ndarray): the real component
		imag (np.ndarray): the imaginary component

	Result:
		np.ndarray: the modulus
		np.ndarray: the argument
	'''
	p = cp if type(real) is cp.ndarray else np
	mod = p.sqrt(real*real + imag*imag)
	arg = p.arctan2(imag, real)
	return mod, arg

@cr('math.vandermonde')
def vandermonde(real:np.ndarray, imag:np.ndarray, m:int, n:int) -> np.ndarray:
	r'''
	Builds a Vandermonde matrix of (m x n) with the real and
	imaginary parts of the eigenvalues

	Args:
		real (np.ndarray): the real component
		imag (np.ndarray): the imaginary component
		m (int): number of rows for the Vandermode matrix
		n (int): number of columns for the Vandermode matrix

	Result:
		np.ndarray: the Vandermonde matrix
	'''
	p = cp if type(real) is cp.ndarray else np
	dtype = p.complex128 if real.dtype is p.double else p.complex64
	Vand  = p.zeros((m, n), dtype=dtype)
	for icol in range(n):
		Vand[:, icol] = (real + imag*1j)**icol
	return Vand

@cr('math.vandermondeTime')
def vandermondeTime(real:np.ndarray, imag:np.ndarray, m:int, time:np.ndarray) -> np.ndarray:
	r'''
	Builds a Vandermonde matrix of (m x n) with the real and
	imaginary parts of the eigenvalues


	Args:
		real (np.ndarray): the real component
		imag (np.ndarray): the imaginary component
		m (int): number of rows for the Vandermode matrix
		time (np.ndarray): the time vector

	Result:
		np.ndarray: the Vandermonde matrix
	'''
	p = cp if type(real) is cp.ndarray else np
	dtype = p.complex128 if real.dtype is p.double else p.complex64
	n = time.shape[0]
	Vand  = p.zeros((m, n), dtype=dtype)
	for it, t in enumerate(time):
		Vand[:, it] = (real + imag*1j)**t
	return Vand

@cr('math.cholesky')
def cholesky(A:np.ndarray) -> np.ndarray:
	r'''
	Returns the Cholesky decompositon of A

	Args:
		A (np.ndarray): Matrix A (M,N)
	
	Result:
		np.ndarray: Cholesky factorization of A (M,N)
	'''
	p = cp if type(A) is cp.ndarray else np
	return p.linalg.cholesky(A)

@cr('math.conj')
def conj(A:np.ndarray) -> np.ndarray:
	r'''
	Conjugates complex number A

	Args:
		A (np.ndarray): Vector, matrix or number A
	
	Result:
		np.ndarray: Conjugate of A
	'''
	p = cp if type(A) is cp.ndarray else np
	return p.conj(A)

@cr('math.inv')
def inv(A:np.ndarray) -> np.ndarray:
	r'''
	Computes the inverse matrix of A

	Args:
		A (np.ndarray): Matrix A (M,N)
	
	Result:
		np.ndarray: Inverse of A (M,N)
	'''
	p = cp if type(A) is cp.ndarray else np
	return p.linalg.inv(A)

@cr('math.flip')
def flip(A:np.ndarray) -> np.ndarray:
	r'''
	Returns the pointwise conjugate of A

	.. warning::
	This function is not implemented in the
	compiled layer and will raise an error if used

	Args:
		A (np.ndarray): Matrix A (M,N)
	
	Result:
		np.ndarray: Flipped version of A (M,N)
	'''
	p = cp if type(A) is cp.ndarray else np
	return p.flip(A)


#!/usr/bin/env python
#
# SPOD analysis.
#
# Last revision: 29/10/2021
from __future__ import print_function, division

import mpi4py
mpi4py.rc.recv_mprobe = False

import numpy as np
import pyLOM


## Data loading
DATAFILE  = './channel.h5'
VARIABLES = ['PRESS','VELOX','VELOY','VELOZ']

m     = pyLOM.Mesh.load(DATAFILE)
d     = pyLOM.Dataset.load(DATAFILE,ptable=m.partition_table)
X     = d.X(*VARIABLES)
npwin = 100 # Number of snapshots in each window
nolap = 20  # Number of overlapping snapshots between windows


## Run SPOD
L, P, f = pyLOM.SPOD.run(X,t,nDFT=npwin,nolap=nolap,remove_mean=True)
pyLOM.SPOD.save('results.h5',L,P,f,d.partition_table,nvars=4,pointData=d.point)
if pyLOM.utils.is_rank_or_serial(root=0): 
	fig,_ = pyLOM.SPOD.plotSpectra(f, L)
	fig.savefig('spectra.png',dpi=300)


## Dump to ParaView
# Spatial modes
modes = np.arange(1,10+1,dtype=np.int32)
d.add_field('spatial_modes_P',len(modes),pyLOM.SPOD.extract_modes(L,P,1,len(d),modes=modes))
d.add_field('spatial_modes_U',len(modes),pyLOM.SPOD.extract_modes(L,P,2,len(d),modes=modes))
d.add_field('spatial_modes_V',len(modes),pyLOM.SPOD.extract_modes(L,P,3,len(d),modes=modes))
d.add_field('spatial_modes_W',len(modes),pyLOM.SPOD.extract_modes(L,P,4,len(d),modes=modes))
pyLOM.io.pv_writer(m,d,'modes',basedir='modes',instants=[0],times=[0.],vars=['spatial_modes_P','spatial_modes_U','spatial_modes_V','spatial_modes_W'],fmt='vtkh5')


## Show and print timings
pyLOM.cr_info()


import numpy as np
from .. import pprint
from ..utils import raiseError

class RegressionEvaluator():
    r"""
    Evaluator class for regression tasks. 
    Includes methods to calculate the mean squared error (MSE), mean absolute error (MAE), 
    mean relative error (MRE), quantiles of the absolute errors, L2 error, and R-squared.

    Args:
        tolerance (float): Tolerance level to consider values close to zero for MRE calculation (default: ``1e-4``).
    """

    def __init__(
        self,
        tolerance: float = 1e-4,
    ) -> None:
        self.tolerance = tolerance

    @property
    def tolerance(self) -> float:
        return self._tolerance
    
    @tolerance.setter
    def tolerance(self, value: float) -> None:
        if value < 0:
            raise raiseError("Tolerance must be a positive value.")
        if value == 0:
            raise raiseError("Tolerance cannot be zero.")
        if value > 1e-2:
            raise raiseError("Tolerance should be less than 1e-2.")
        self._tolerance = value

    def mean_squared_error(
        self,
        y_true: np.ndarray,
        y_pred: np.ndarray,
    ) -> float:
        """
        Compute the mean squared error (MSE) between the true values and the predicted values.

        Args:
            y_true (numpy.ndarray): The true values.
            y_pred (numpy.ndarray): The predicted values.

        Returns:
            float: The mean squared error.
        """
        return np.mean((y_true - y_pred) ** 2)
    
    def mean_absolute_error(self, y_true, y_pred):
        """
        Compute the mean absolute error (MAE) between the true values and the predicted values.

        Args:
            y_true (numpy.ndarray): The true values.
            y_pred (numpy.ndarray): The predicted values.

        Returns:
            float: The mean absolute error.
        """
        return np.mean(np.abs(y_true - y_pred))
    
    def mean_relative_error(
        self,
        y_pred: np.ndarray,
        y_true: np.ndarray,
    ) -> float:
        """
        Compute the mean relative error (MRE) between the true values and the predicted values,
        adding a tolerance level to consider values close to zero.

        Args:
            y_true (numpy.ndarray): The true values.
            y_pred (numpy.ndarray): The predicted values.
            tolerance (float): Tolerance level to consider values close to zero. Default is 1e-4.

        Returns:
            float: The mean relative error excluding cases where y_true is close to zero.
        """
        relative_errors = np.abs((y_true - y_pred) / (y_true + self.tolerance))
        return np.mean(relative_errors) * 100
    
    def ae_q(
        self,
        y_pred: np.ndarray,
        y_true: np.ndarray,
        quantile: int,
    ) -> float:
        """
        Calculate the quantile of the absolute errors between the true and predicted values.

        Args:
            y_true (numpy.ndarray): The true values.
            y_pred (numpy.ndarray): The predicted values.
            quantile (int): The quantile to calculate. Must be between 0 and 100.

        Returns:
            float: The quantile of the absolute errors.
        """

        absolute_errors = np.abs(y_true - y_pred)
        return np.percentile(absolute_errors, quantile)    

    def l2_error(
        self,
        y_pred: np.ndarray,
        y_true: np.ndarray,
    ) -> float:
        """
        Calculate the L2 error between the true and predicted values.

        Args:
            y_true (numpy.ndarray): The true values.
            y_pred (numpy.ndarray): The predicted values.

        Returns:
            float: The L2 error.
        """

        return np.linalg.norm(y_true - y_pred) / np.linalg.norm(y_true)
    
    def R2(
        self,
        y_true: np.ndarray,
        y_pred: np.ndarray,
    ) -> float:
        """
        Calculate the R-squared (coefficient of determination) for a set of true and predicted values.

        Args:
            y_true (numpy.ndarray): The true values.
            y_pred (numpy.ndarray): The predicted values.

        Returns:
            float: The R-squared value.
        """
        y_mean = np.mean(y_true)
        total_sum_of_squares = np.sum((y_true - y_mean) ** 2)
        residual_sum_of_squares = np.sum((y_true - y_pred) ** 2)
        r_squared = 1 - (residual_sum_of_squares / total_sum_of_squares)
        return r_squared

    def print_metrics(self):
        """
        Print the calculated regression metrics.
        """
        if self._metrics is None:
            raise raiseError("No metrics have been calculated yet.")
        
        pprint(0, "\nRegression evaluator metrics:")
        
        for key, value in self._metrics.items():
            if key == "mre":
                pprint(0, f"{key}: {value:.4f}%")
            elif key == "r2":
                pprint(0, f"{key}: {value:.4f}")
            else:
                if value < 1e-3 or value > 1e3:
                    pprint(0, f"{key}: {value:.4e}")
                else:
                    pprint(0, f"{key}: {value:.4f}")


    def __call__(
        self,
        y_true: np.ndarray,
        y_pred: np.ndarray,
    ) -> dict:
        """
        Calculate multiple regression metrics between the true and predicted values.

        Args:
            y_true (numpy.ndarray): An array-like object containing the true values.
            y_pred (numpy.ndarray): An array-like object containing the predicted values.

        Returns:
            dict: A dictionary containing the calculated regression metrics.
        """
        try:
            y_true = np.array(y_true)
            y_pred = np.array(y_pred)
        except AttributeError:
            raise raiseError(f"could not create numpy arrays from object with type {type(y_true)}")
        
        mse = self.mean_squared_error(y_true, y_pred)
        rmse = np.sqrt(mse)
        mae = self.mean_absolute_error(y_true, y_pred)
        mre = self.mean_relative_error(y_true, y_pred)
        aq_95 = self.ae_q(y_true, y_pred, 95)
        aq_99 = self.ae_q(y_true, y_pred, 99)
        r2 = self.R2(y_true, y_pred)
        l2_error = self.l2_error(y_true, y_pred)
        self._metrics = {
            "mse": mse,
            "rmse": rmse,
            "mae": mae, 
            "mre": mre, 
            "ae_95": aq_95,
            "ae_99": aq_99,
            "r2": r2,
            "l2_error": l2_error 
        }
        return self._metrics


#!/usr/bin/env python
#
# pyLOM, setup.
#
# Setup and cythonize code.
#
# Last rev: 28/10/2021
from __future__ import print_function, division

import os, sys, numpy as np, mpi4py
from setuptools import setup, Extension, find_packages
from Cython.Build import cythonize


## Read INIT file
with open('pyLOM/__init__.py') as f:
	for l in f.readlines():
		if '__version__' in l:
			__version__ = eval(l.split('=')[1].strip())


## Read README file
with open('README.md') as f:
	readme = f.read()


## Read compilation options
options = {}
with open('options.cfg') as f:
	for line in f.readlines():
		if '#' in line or len(line) == 1: continue # Skip comment
		linep = line.split('=')
		options[linep[0].strip()] = linep[1].strip()
		if options[linep[0].strip()] == 'ON':  options[linep[0].strip()] = True
		if options[linep[0].strip()] == 'OFF': options[linep[0].strip()] = False
options['MODULES_COMPILED'] = options['MODULES_COMPILED'].lower().split(',')


## Set up compiler options and flags
ICC = 'icx' if 'ACC' in options['PLATFORM'] else 'icc'
CC  = 'mpicc'   if options['FORCE_GCC'] or not os.system('which %s > /dev/null'%ICC) == 0 else 'mpiicc'
CXX = 'mpicxx'  if options['FORCE_GCC'] or not os.system('which %s > /dev/null'%ICC) == 0 else 'mpiicpc'
FC  = 'mpifort' if options['FORCE_GCC'] or not os.system('which %s > /dev/null'%ICC) == 0 else 'mpiifort'

CFLAGS   = ''
CXXFLAGS = ' -std=c++11'
FFLAGS   = ''
DFLAGS   = ' -DNPY_NO_DEPRECATED_API'
if options['USE_MKL']:   DFLAGS += ' -DUSE_MKL'
if options['USE_FFTW']:  DFLAGS += ' -DUSE_FFTW3'
if options['USE_GESVD']: DFLAGS += ' -DUSE_LAPACK_DGESVD'
if CC == 'mpicc':
	# Using GCC as a compiler
	CFLAGS   += ' -O0 -g -rdynamic -fPIC' if options['DEBUGGING'] else ' -O%s -ffast-math -fPIC' % options['OPTL']
	CXXFLAGS += ' -O0 -g -rdynamic -fPIC' if options['DEBUGGING'] else ' -O%s -ffast-math -fPIC' % options['OPTL']
	FFLAGS   += ' -O0 -g -rdynamic -fPIC' if options['DEBUGGING'] else ' -O%s -ffast-math -fPIC' % options['OPTL']
	# Vectorization flags
	if options['VECTORIZATION']:
		CFLAGS   += ' -march=native -ftree-vectorize'
		CXXFLAGS += ' -march=native -ftree-vectorize'
		FFLAGS   += ' -march=native -ftree-vectorize'
	# OpenMP flag
	if options['OPENMP_PARALL']:
		CFLAGS   += ' -fopenmp'
		CXXFLAGS += ' -fopenmp'
else:
	# Using GCC as a compiler
	CFLAGS   += ' -O0 -g -traceback -fPIC' if options['DEBUGGING'] else ' -O%s -fPIC' % options['OPTL']
	CXXFLAGS += ' -O0 -g -traceback -fPIC' if options['DEBUGGING'] else ' -O%s -fPIC' % options['OPTL']
	FFLAGS   += ' -O0 -g -traceback -fPIC' if options['DEBUGGING'] else ' -O%s -fPIC' % options['OPTL']
	# Vectorization flags
	if options['VECTORIZATION']:
		CFLAGS   += ' -x%s -mtune=%s' % (options['HOST'],options['TUNE'])
		CXXFLAGS += ' -x%s -mtune=%s' % (options['HOST'],options['TUNE'])
		FFLAGS   += ' -x%s -mtune=%s' % (options['HOST'],options['TUNE'])
	# OpenMP flag
	if options['OPENMP_PARALL']:
		CFLAGS   += ' -qopenmp'
		CXXFLAGS += ' -qopenmp'
		DFLAGS   += ' -DUSE_OMP'


## Set up environment variables
os.environ['CC']       = CC
os.environ['CXX']      = CXX
os.environ['CFLAGS']   = CFLAGS + DFLAGS
os.environ['CXXFLAGS'] = CXXFLAGS + DFLAGS
os.environ['LDSHARED'] = CC + ' -shared'


## Libraries and includes
include_dirs  = []
extra_objects = []
libraries     = ['m']

# OSX needs to also link with python for reasons...
if sys.platform == 'darwin': libraries += [f'python{sys.version_info[0]}.{sys.version_info[1]}']

if options['USE_FFTW']:
	# NFFT
	include_dirs  += ['Deps/nfft/include']
	extra_objects += ['Deps/nfft/lib/libnfft3.a','Deps/nfft/lib/libnfft3f.a']
	# FFTW
	include_dirs  += ['Deps/fftw/include']
	extra_objects += ['Deps/fftw/lib/libfftw3.a','Deps/fftw/lib/libfftw3f.a']
	if options['OPENMP_PARALL']: extra_objects += ['Deps/fftw/lib/libfftw3_omp.a','Deps/fftw/lib/libfftw3f_omp.a']
	DFLAGS        += ' -DUSE_FFTW3'
else:
	# KISSFFT
	include_dirs  += ['Deps/kissfft/include/kissfft']
	extra_objects += ['Deps/kissfft/lib/libkissfft-double.a']

# Select which libraries to use depending on the compilation options
if options['USE_MKL']:
	# Link with Intel MKL using the intel compilers
	# this is the most performing option available
	mklroot        = 'Deps/oneAPI/mkl'
	include_dirs  += [os.path.join(mklroot,'include')]
	if options['OPENMP_PARALL']:
		extra_objects += [os.path.join(mklroot,'libmkl_intel_thread.a' if CC == 'mpiicc' else 'libmkl_gcc_thread.a')]
	else:
		extra_objects += [os.path.join(mklroot,'libmkl_intel.a' if CC == 'mpiicc' else 'libmkl_gcc.a')]
else:
	# Link with OpenBLAS which has a decent performance but is not
	# as fast as Intel MKL
	include_dirs  += ['Deps/lapack/include/openblas']
	extra_objects += ['Deps/lapack/lib/libopenblas.a']
	libraries     += ['gfortran']
	# Classical LAPACK & BLAS library has a very bad performance
	# but is left here for nostalgia
#	include_dirs  += ['Deps/lapack/include/']
#	extra_objects += ['Deps/lapack/lib/liblapacke.a','Deps/lapack/lib/liblapack.a','Deps/lapack/lib/libcblas.a','Deps/lapack/lib/libblas.a']
#	libraries     += ['gfortran']


## Modules
# vmmath module
Module_cfuncs     = Extension('pyLOM.vmmath.cfuncs',
						sources       = ['pyLOM/vmmath/cfuncs.pyx',
										 'pyLOM/vmmath/src/vector_matrix.c',
										 'pyLOM/vmmath/src/averaging.c',
										 'pyLOM/vmmath/src/svd.c',
										 'pyLOM/vmmath/src/fft.c',
										 'pyLOM/vmmath/src/geometric.c',
										 'pyLOM/vmmath/src/truncation.c',
										 'pyLOM/vmmath/src/stats.c',
										 'pyLOM/vmmath/src/regression.c',
									    ],
						language      = 'c',
						include_dirs  = include_dirs + ['pyLOM/vmmath/src',np.get_include(),mpi4py.get_include()],
						extra_objects = extra_objects,
						libraries     = libraries,
					   )
Module_maths     = Extension('pyLOM.vmmath.maths',
						sources       = ['pyLOM/vmmath/maths.pyx',
										 'pyLOM/vmmath/src/vector_matrix.c',
									    ],
						language      = 'c',
						include_dirs  = include_dirs + ['pyLOM/vmmath/src',np.get_include(),mpi4py.get_include()],
						extra_objects = extra_objects,
						libraries     = libraries,
					   )
Module_averaging = Extension('pyLOM.vmmath.averaging',
						sources       = ['pyLOM/vmmath/averaging.pyx',
										 'pyLOM/vmmath/src/averaging.c',
									    ],
						language      = 'c',
						include_dirs  = include_dirs + ['pyLOM/vmmath/src',np.get_include(),mpi4py.get_include()],
						extra_objects = extra_objects,
						libraries     = libraries,
					   )
Module_svd       = Extension('pyLOM.vmmath.svd',
						sources       = ['pyLOM/vmmath/svd.pyx',
										 'pyLOM/vmmath/src/vector_matrix.c',
										 'pyLOM/vmmath/src/svd.c',
									    ],
						language      = 'c',
						include_dirs  = include_dirs + ['pyLOM/vmmath/src',np.get_include(),mpi4py.get_include()],
						extra_objects = extra_objects,
						libraries     = libraries,
					   )
Module_fft       = Extension('pyLOM.vmmath.fft',
						sources       = ['pyLOM/vmmath/fft.pyx',
										 'pyLOM/vmmath/src/fft.c',
									    ],
						language      = 'c',
						include_dirs  = include_dirs + ['pyLOM/vmmath/src',np.get_include(),mpi4py.get_include()],
						extra_objects = extra_objects,
						libraries     = libraries,
					   )
Module_geometric = Extension('pyLOM.vmmath.geometric',
						sources       = ['pyLOM/vmmath/geometric.pyx',
										 'pyLOM/vmmath/src/geometric.c',
									    ],
						language      = 'c',
						include_dirs  = include_dirs + ['pyLOM/vmmath/src',np.get_include(),mpi4py.get_include()],
						extra_objects = extra_objects,
						libraries     = libraries,
					   )
Module_truncation = Extension('pyLOM.vmmath.truncation',
						sources       = ['pyLOM/vmmath/truncation.pyx',
										 'pyLOM/vmmath/src/vector_matrix.c',
										 'pyLOM/vmmath/src/truncation.c',
									    ],
						language      = 'c',
						include_dirs  = include_dirs + ['pyLOM/vmmath/src',np.get_include(),mpi4py.get_include()],
						extra_objects = extra_objects,
						libraries     = libraries,
					   )
Module_stats     = Extension('pyLOM.vmmath.stats',
						sources       = ['pyLOM/vmmath/stats.pyx',
										 'pyLOM/vmmath/src/stats.c',
									    ],
						language      = 'c',
						include_dirs  = include_dirs + ['pyLOM/vmmath/src',np.get_include(),mpi4py.get_include()],
						extra_objects = extra_objects,
						libraries     = libraries,
					   )
Module_regression = Extension('pyLOM.vmmath.regression',
						sources       = ['pyLOM/vmmath/regression.pyx',
										 'pyLOM/vmmath/src/vector_matrix.c',
										 'pyLOM/vmmath/src/regression.c',
									    ],
						language      = 'c',
						include_dirs  = include_dirs + ['pyLOM/vmmath/src',np.get_include(),mpi4py.get_include()],
						extra_objects = extra_objects,
						libraries     = libraries,
					   )
# low-order modules
Module_POD = Extension('pyLOM.POD.wrapper',
						sources       = ['pyLOM/POD/wrapper.pyx',
										 'pyLOM/vmmath/src/vector_matrix.c',
										 'pyLOM/vmmath/src/averaging.c',
										 'pyLOM/vmmath/src/svd.c',
										 'pyLOM/vmmath/src/truncation.c',
									    ],
						language      = 'c',
						include_dirs  = include_dirs + ['pyLOM/vmmath/src',np.get_include(),mpi4py.get_include()],
						extra_objects = extra_objects,
						libraries     = libraries,
					   )
Module_DMD = Extension('pyLOM.DMD.wrapper',
						sources       = ['pyLOM/DMD/wrapper.pyx',
										 'pyLOM/vmmath/src/vector_matrix.c',
										 'pyLOM/vmmath/src/averaging.c',
										 'pyLOM/vmmath/src/svd.c',
										 'pyLOM/vmmath/src/truncation.c',
									    ],
						language      = 'c',
						include_dirs  = include_dirs + ['pyLOM/vmmath/src',np.get_include(),mpi4py.get_include()],
						extra_objects = extra_objects,
						libraries     = libraries,
					   )
Module_SPOD = Extension('pyLOM.SPOD.wrapper',
						sources       = ['pyLOM/SPOD/wrapper.pyx',
										 'pyLOM/vmmath/src/vector_matrix.c',
										 'pyLOM/vmmath/src/averaging.c',
										 'pyLOM/vmmath/src/svd.c',
										 'pyLOM/vmmath/src/fft.c',
									    ],
						language      = 'c',
						include_dirs  = include_dirs + ['pyLOM/vmmath/src',np.get_include(),mpi4py.get_include()],
						extra_objects = extra_objects,
						libraries     = libraries,
					   )


## Build modules
# Math module
Module_Math  = [Module_cfuncs]
Module_Math += [Module_maths]      if 'math.maths'      in options['MODULES_COMPILED'] else []
Module_Math += [Module_averaging]  if 'math.averaging'  in options['MODULES_COMPILED'] else []
Module_Math += [Module_svd]        if 'math.svd'        in options['MODULES_COMPILED'] else []
Module_Math += [Module_fft]        if 'math.fft'        in options['MODULES_COMPILED'] else []
Module_Math += [Module_geometric]  if 'math.geometric'  in options['MODULES_COMPILED'] else []
Module_Math += [Module_truncation] if 'math.truncation' in options['MODULES_COMPILED'] else []
Module_Math += [Module_stats]      if 'math.stats'      in options['MODULES_COMPILED'] else []
Module_Math += [Module_regression] if 'math.regression' in options['MODULES_COMPILED'] else []
# ROM module
Module_ROM   = [Module_POD]  if 'rom.pod'  in options['MODULES_COMPILED'] else []
Module_ROM  += [Module_DMD]  if 'rom.dmd'  in options['MODULES_COMPILED'] else []
Module_ROM  += [Module_SPOD] if 'rom.spod' in options['MODULES_COMPILED'] else []


## Decide which modules to compile
modules_list = Module_Math + Module_ROM if options['USE_COMPILED'] else []


## Main setup
setup(
	name             = 'pyLowOrder',
	version          = __version__,
	author           = 'Benet Eiximeno, Beka Begiashvili, Arnau Miro, Eusebio Valero, Oriol Lehmkuhl',
	author_email     = 'benet.eiximeno@bsc.es, beka.begiashvili@alumnos.upm.es, arnau.mirojane@bsc.es, eusebio.valero@upm.es, oriol.lehmkuhl@bsc.es',
	maintainer       = 'Benet Eiximeno, Arnau Miro',
	maintainer_email = 'benet.eiximeno@bsc.es, arnau.mirojane@bsc.es',
	ext_modules      = cythonize(modules_list,
		language_level = str(sys.version_info[0]), # This is to specify python 3 synthax
		annotate       = True                      # This is to generate a report on the conversion to C code
	),
    long_description = readme,
    url              = 'https://github.com/ArnauMiro/pyLowOrder',
    packages         = find_packages(exclude=('Converters','Examples','Deps','Testsuite','Tools')),
	install_requires = ['numpy','matplotlib','cython>=3.0.0','h5py>=3.0.0','mpi4py>=4.0.0','nfft']
)



#!/usr/bin/env python
#
# pyLOM - Python Low Order Modeling.
#
# NN callback routines.
#
# Last rev: 02/10/2024
from __future__ import print_function

import numpy as np


## Early stopper callback
class EarlyStopper:
    r"""
    Early stopper callback.

    Args:
        patience (int): Number of epochs to wait before stopping the training. Default: ``1``.
        min_delta (float): Minimum change in the monitored quantity to qualify as an improvement. Default: ``0``.
    """
    def __init__(self, patience: int = 1, min_delta: float = 0):
        self.patience = patience
        self.min_delta = min_delta
        self.counter = 0
        self.min_validation_loss = np.inf
       
    def early_stop(self, validation_loss: float, prev_train: float, train: float) -> bool:
        r"""
        Early stopper routine. The training will stop if the validation loss does not improve for a number of epochs.

        Args:
            validation_loss (float): Validation loss.
            prev_train (float): Previous training loss.
            train (float): Current training loss.

        Returns:
            bool: True if early stopping is triggered, False otherwise.
        """
        if validation_loss < self.min_validation_loss:
            self.min_validation_loss = validation_loss
            self.counter = 0
        elif validation_loss > (self.min_validation_loss + self.min_delta):
            self.counter += 1
            if self.counter >= self.patience:
                return True
        elif prev_train < train:
            self.counter += 1
            if self.counter >= self.patience:
                return True
        return False